/*  avwcomplex.cc

    Mark Jenkinson, FMRIB Image Analysis Group

    Copyright (C) 2000-2001 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

#include "newimageall.h"

using namespace NEWIMAGE;

void print_usage(const string& progname) {
  cerr << "Usage: " << progname << " <-outputtype> <input> <output> [startvol [endvol]]" << endl << endl;

  cerr << "       " << progname << " -realabs complexvol absvol" 
       << " [startvol [endvol]]" << endl;
  cerr << "       " << progname << " -realphase complexvol phasevol" 
       << " [startvol [endvol]]" << endl;
  cerr << "       " << progname << " -realpolar complexvol absvol phasevol" 
       << " [startvol [endvol]]" << endl;
  cerr << "       " << progname << " -realcartesian complexvol realvol"
       << " imagvol [startvol [endvol]]" << endl;
  cerr << "       " << progname << " -complex realvol imagvol" 
       << " complexvol [startvol [endvol]]" << endl;
  cerr << "       " << progname << " -complexpolar absvol phasevol" 
       << " complexvol [startvol [endvol]]" << endl;
  cerr << "       " << progname << " -complexsplit source dest" 
       << " [startvol [endvol]]" << endl;
  cerr << "       " << progname << " -complexmerge source1 source2 dest" 
       << endl;
}

void fix_start_and_end(int& start, int& end, int mint, int maxt)
{
  if (start<0) { start=mint; }
  if (start<mint) start=mint;
  if (start>maxt) start=maxt;
  if (end<0) { end=maxt; }
  if (end<mint) end=mint;
  if (end>maxt) end=maxt;
}


void realpolar(const string& fin, const string& fabs, const string& fphase, 
	       int start, int end) 
{
  volume4D<float> vreal, vimag;
  volumeinfo vinfo;
  read_complexvolume4D(vreal, vimag, fin, vinfo);
  fix_start_and_end(start,end,vreal.mint(),vreal.maxt());

  if (fabs.size()>0) {
    volume4D<float> vabs;
    for (int n=start; n<=end; n++) {
      vabs.addvolume(abs(vreal[n],vimag[n]));
    }
    save_volume4D(vabs,fabs,vinfo);
  }

  if (fphase.size()>0) {
    volume4D<float> vphase;
    for (int n=start; n<=end; n++) {
      vphase.addvolume(phase(vreal[n],vimag[n]));
    }
    save_volume4D(vphase,fphase,vinfo);
  }
}

void realcartesian(const string& fin, const string& freal, const string& fimag, 
		   int start, int end)
{
  volume4D<float> vreal, vimag;
  volumeinfo vinfo;
  read_complexvolume4D(vreal, vimag, fin, vinfo);
  fix_start_and_end(start,end,vreal.mint(),vreal.maxt());

  if (freal.size()>0) {
    volume4D<float> vre;
    for (int n=start; n<=end; n++) {
      vre.addvolume(vreal[n]);
    }
    save_volume4D(vre,freal,vinfo);
  }

  if (fimag.size()>0) {
    volume4D<float> vim;
    for (int n=start; n<=end; n++) {
      vim.addvolume(vimag[n]);
    }
    save_volume4D(vim,fimag,vinfo);
  }
}


void complexsave(const string& freal, const string& fimag, 
		 const string& fcomplex, int start, int end)
{
  volume4D<float> vreal, vimag;
  volumeinfo vinfo;
  read_volume4D(vreal, freal, vinfo);
  read_volume4D(vimag, fimag, vinfo);
  fix_start_and_end(start,end,Max(vreal.mint(),vimag.mint()),
		    Min(vreal.maxt(),vimag.maxt()));

  if (fcomplex.size()>0) {
    volume4D<float> vre, vim;
    for (int n=start; n<=end; n++) {
      vre.addvolume(vreal[n]);
      vim.addvolume(vimag[n]);
    }
    save_complexvolume4D(vre,vim,fcomplex,vinfo);
  }
}


void complexpolar(const string& fabsvol, const string& fphasevol, 
		  const string& fcomplex, int start, int end)
{
  volume4D<float> vabs, vphase;
  volumeinfo vinfo;
  read_volume4D(vabs, fabsvol, vinfo);
  read_volume4D(vphase, fphasevol, vinfo);
  fix_start_and_end(start,end,Max(vabs.mint(),vphase.mint()),
		    Min(vabs.maxt(),vphase.maxt()));

  if (fcomplex.size()>0) {
    volume4D<float> vre, vim;
    for (int n=start; n<=end; n++) {
      vre.addvolume(real(vabs[n],vphase[n]));
      vim.addvolume(imag(vabs[n],vphase[n]));
    }
    save_complexvolume4D(vre,vim,fcomplex,vinfo);
  }
}

void complexsplit(const string& fsource, const string& fdest, 
		  int start, int end)
{
  volume4D<float> vreal, vimag;
  volumeinfo vinfo;
  read_complexvolume4D(vreal, vimag, fsource, vinfo);
  fix_start_and_end(start,end,Max(vreal.mint(),vimag.mint()),
		    Min(vreal.maxt(),vimag.maxt()));

  if (fdest.size()>0) {
    volume4D<float> vre, vim;
    for (int n=start; n<=end; n++) {
      vre.addvolume(vreal[n]);
      vim.addvolume(vimag[n]);
    }
    save_complexvolume4D(vre,vim,fdest,vinfo);
  }
}

void complexmerge(const string& fsource1, const string& fsource2,
		  const string& fdest)
{
  volume4D<float> vr1, vi1, vr2, vi2;
  volumeinfo vinfo;
  read_complexvolume4D(vr1, vi1, fsource1, vinfo);
  read_complexvolume4D(vr2, vi2, fsource2, vinfo);

  if (!samesize(vr1,vi1)) {
    cerr << "Could not process image " << fsource1 << " correctly." << endl;
    return;
  }
  if (!samesize(vr2,vi2)) {
    cerr << "Could not process image " << fsource2 << " correctly." << endl;
    return;
  }
  if (fdest.size()>0) {
    volume4D<float> vdr, vdi;
    for (int n=vr1.mint(); n<=vr1.maxt(); n++) {
      vdr.addvolume(vr1[n]);
      vdr.addvolume(vi1[n]);
    }
    for (int n=vr2.mint(); n<=vr2.maxt(); n++) {
      vdr.addvolume(vr2[n]);
      vdr.addvolume(vi2[n]);
    }
    save_complexvolume4D(vdr,vdi,fdest,vinfo);
  }
}


int main(int argc,char *argv[])
{

  Tracer tr("main");

  string progname=argv[0];
  if (argc<4) { 
    print_usage(progname);
    return -1; 
  }

  string arg = argv[1];
  int start=-1, end=-1;

  if (arg=="-realabs") {
    if (argc<4) { print_usage(progname); return -1; }
    if (argc>=5)  start=atoi(argv[4]);
    if (argc>=6)  end=atoi(argv[5]);
    realpolar(argv[2],argv[3],"",start,end);
  } else if (arg=="-realphase") {
    if (argc<4) { print_usage(progname); return -1; }
    if (argc>=5)  start=atoi(argv[4]);
    if (argc>=6)  end=atoi(argv[5]);
    realpolar(argv[2],"",argv[3],start,end);
  } else if (arg=="-realpolar") {
    if (argc<5) { print_usage(progname); return -1; }
    if (argc>=6)  start=atoi(argv[5]);
    if (argc>=7)  end=atoi(argv[6]);
    realpolar(argv[2],argv[3],argv[4],start,end);
  } else if (arg=="-realcartesian") {
    if (argc<5) { print_usage(progname); return -1; }
    if (argc>=6)  start=atoi(argv[5]);
    if (argc>=7)  end=atoi(argv[6]);
    realcartesian(argv[2],argv[3],argv[4],start,end);
  } else if (arg=="-complex") {
    if (argc<5) { print_usage(progname); return -1; }
    if (argc>=6)  start=atoi(argv[5]);
    if (argc>=7)  end=atoi(argv[6]);
    complexsave(argv[2],argv[3],argv[4],start,end);
  } else if (arg=="-complexpolar") {
    if (argc<5) { print_usage(progname); return -1; }
    if (argc>=6)  start=atoi(argv[5]);
    if (argc>=7)  end=atoi(argv[6]);
    complexpolar(argv[2],argv[3],argv[4],start,end);
  } else if (arg=="-complexsplit") {
    if (argc<4) { print_usage(progname); return -1; }
    if (argc>=5)  start=atoi(argv[4]);
    if (argc>=6)  end=atoi(argv[5]);
    complexsplit(argv[2],argv[3],start,end);
  } else if (arg=="-complexmerge") {
    if (argc<5) { print_usage(progname); return -1; }
    complexmerge(argv[2],argv[3],argv[4]);
  } else {
    print_usage(progname); return -1;
  }

  return 0;
}

