/* {{{ Copyright etc. */

/*  feat_model - create FEAT design matrix and contrast file

    Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 1999-2001 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

/* }}} */
/* {{{ includes and defines */

#include "libss/libss.h"

/* }}} */
/* {{{ numerical algorithms */

/* {{{ gammpdf */

float gammpdf(float h, float l, float x)
{
        int i;
        double W=h+5.5,X=h,Y=h,Z=1.000000000190015,gammaln;
        static double cof[6]={76.18009172947146,-86.50532032941677,
                24.01409824083091,-1.231739572450155,
                0.1208650973866179e-2,-0.5395239384953e-5};

	if (x<=0.0 || h<=0.0 || l<=0.0) return 0.0;

        W -= (X+0.5)*log(W);
        Z=1.000000000190015;
        for (i=0;i<=5;i++) Z += cof[i]/++Y;
        gammaln=-W+log(2.5066282746310005*Z/X);

	/* equation is   pow(l,h) * pow(x,h-1) * exp(-l*x - gammln)  but replace with more numerically sensible equivalent */
        return exp( h*log(l) + (h-1)*log(x) - l*x - gammaln );
}

/* }}} */
/* {{{ freeSVD */

/* This SVD routine is based on pgs 30-48 of "Compact Numerical Methods
   for Computers" by J.C. Nash (1990), used to compute the pseudoinverse.
   Modifications include:
        Translation from Pascal to ANSI C.
        Array indexing from 0 rather than 1.
        Float replaced by double everywhere.
        Support for the Matrix structure.
        I changed the array indexing so that the matricies (float [][])
           could be replaced be a single list (double *) for more
           efficient communication with Mathematica.
*/

#define TOLERANCE 1.0e-22

void SVD(double *W, double *Z, int nRow, int nCol)
{
  int i, j, k, EstColRank, RotCount, SweepCount, slimit;
  double eps, e2, tol, vt, p, x0, y0, q, r, c0, s0, d1, d2;
  eps = TOLERANCE;
  slimit = nCol/4;
  if (slimit < 6.0)
    slimit = 6;
  SweepCount = 0;
  e2 = 10.0*nRow*eps*eps;
  tol = eps*.1;
  EstColRank = nCol;
  for (i=0; i<nCol; i++)
    for (j=0; j<nCol; j++)
      {
        W[nCol*(nRow+i)+j] = 0.0;
        W[nCol*(nRow+i)+i] = 1.0;
      }
  RotCount = EstColRank*(EstColRank-1)/2;
  while (RotCount != 0 && SweepCount <= slimit)
    {
      RotCount = EstColRank*(EstColRank-1)/2;
      SweepCount++;
      for (j=0; j<EstColRank-1; j++)
        {
          for (k=j+1; k<EstColRank; k++)
            {
              p = q = r = 0.0;
              for (i=0; i<nRow; i++)
                {
                  x0 = W[nCol*i+j]; y0 = W[nCol*i+k];
                  p += x0*y0; q += x0*x0; r += y0*y0;
                }
              Z[j] = q; Z[k] = r;
              if (q >= r)
                {
                  if (q<=e2*Z[0] || fabs(p)<=tol*q) RotCount--;
                  else
                    {
                      p /= q; r = 1 - r/q; vt = sqrt(4*p*p+r*r);
                      c0 = sqrt(fabs(.5*(1+r/vt))); s0 = p/(vt*c0);
                      for (i=0; i<nRow+nCol; i++)
                        {
                          d1 = W[nCol*i+j]; d2 = W[nCol*i+k];
                          W[nCol*i+j] = d1*c0+d2*s0; W[nCol*i+k] = -d1*s0+d2*c0;
                        }
                    }
                }
              else
                {
                  p /= r; q = q/r-1; vt = sqrt(4*p*p+q*q);
                  s0 = sqrt(fabs(.5*(1-q/vt)));
                  if (p<0) s0 = -s0;
                  c0 = p/(vt*s0);
                  for (i=0; i<nRow+nCol; i++)
                    {
                      d1 = W[nCol*i+j]; d2 = W[nCol*i+k];
                      W[nCol*i+j] = d1*c0+d2*s0; W[nCol*i+k] = -d1*s0+d2*c0;
                    }
                }
            }
        }
      while (EstColRank>=3 &&
	     Z[EstColRank-1]<=Z[0]*tol+tol*tol)
        EstColRank--;
    }
}

/* }}} */

/* }}} */
/* {{{ find_line */

char *find_line(char *filename, char *key, char *fl)
{
  FILE *fd;
  int  j;

  fd=fopen(filename,"r");

  while ( fgets(fl, 1000, fd) != NULL )
    if (strncmp(fl,"set ",4)==0)
      for(j=4; fl[j]!=0; j++)
	if (strncmp(key,fl+j,strlen(key))==0)
	  {
	    fclose(fd);
	    if (fl[strlen(fl)-1]==10) fl[strlen(fl)-1]=0;
	    return(fl+j+1+strlen(key));
	  }

  printf("Can't find key %s\n",key);
  exit(1);
}

/* }}} */
/* {{{ main() */

#define DT 0.05           /* temporal sampling for initial model, in seconds */
#define MINCONDITION 5e-2 /* used to test ratio of min:max eigenvalues after SVD of matrix, for rank testing */

int main(argc,argv)
  int argc;
  char *argv[];
{
  /* {{{ variables */

FILE   *ofp;
int    i, t, con, ev, Xev, npts, ndelete, mnpts, tmp,
  nevs, shape, convolve, deriv_yn, ncon, nftests, *origevs, tempfiltyn,
  red_yn, grey_yn, maxcols,
  constcol=0, temphp_yn, templp_yn, f, *F;
float  *X, *C, tr, dt=DT, mult, trmult, version, nltffwhm, condition, tmpf;
double *W, *Z=NULL, minl, maxl;
char   fn[10000], filename[10000], fl[10000], key[10000];

/* }}} */
  /* {{{ read arguments and prepare main variables */

if (argc<2)
{
  printf("Usage: feat_model <file_root>\n");
  exit(1);
}

sprintf(fn,"%s.fsf",argv[1]);

version=atof(find_line(fn, "fmri(version)", fl));
tr=atof(find_line(fn, "fmri(tr)", fl));
npts=atoi(find_line(fn, "fmri(npts)", fl));
ndelete=atoi(find_line(fn, "fmri(ndelete)", fl));
temphp_yn=atoi(find_line(fn, "fmri(temphp_yn)", fl));
templp_yn=atoi(find_line(fn, "fmri(templp_yn)", fl));
nltffwhm=atof(find_line(fn, "fmri(paradigm_hp)", fl));
nevs=atoi(find_line(fn, "fmri(evs)", fl));
ncon=atoi(find_line(fn, "fmri(ncon)", fl));
nftests=atoi(find_line(fn, "fmri(nftests)", fl));
red_yn=atoi(find_line(fn, "fmri(dmred)", fl));
grey_yn=atoi(find_line(fn, "fmri(dmgrey)", fl));
constcol=atoi(find_line(fn, "fmri(constcol)", fl));

npts -= ndelete;        /* subtract off images which have already been deleted */
mult=1.0/dt;            /* points in high-res model per second */
trmult=tr*mult;         /* points in high-res model per TR */
mnpts=(npts+10)*trmult; /* number of points in high-res model - add a few on for safety */
maxcols=nevs*2+1;       /* maximum possible columns in design matrix - may not all get used */

X=(float *)malloc(sizeof(float)*maxcols*mnpts);
origevs=(int *)calloc(maxcols,sizeof(int));     /* init to zero */

/* }}} */
  /* {{{ read and create design matrix */

for(Xev=ev=0; ev<nevs; ev++, Xev++)
{
  origevs[Xev]=1;

  sprintf(key,"fmri(shape%d)",ev+1); shape=atoi(find_line(fn, key, fl));
  sprintf(key,"fmri(convolve%d)",ev+1); convolve=atoi(find_line(fn, key, fl));
  sprintf(key,"fmri(tempfilt_yn%d)",ev+1); tempfiltyn=atoi(find_line(fn, key, fl));
  sprintf(key,"fmri(deriv_yn%d)",ev+1); deriv_yn=atoi(find_line(fn, key, fl));

  /* {{{ create basic shape */

  for(t=0;t<mnpts;t++)
    X[t*maxcols+Xev]=0.0;

  switch (shape)
    {
    case 0:
      /* {{{ square wave */

{
  float skip, off, on, phase, stop;

  sprintf(key,"fmri(skip%d)",ev+1); skip=atof(find_line(fn, key, fl));
  sprintf(key,"fmri(off%d)",ev+1); off=atof(find_line(fn, key, fl));
  sprintf(key,"fmri(on%d)",ev+1); on=atof(find_line(fn, key, fl));
  sprintf(key,"fmri(phase%d)",ev+1); phase=atof(find_line(fn, key, fl));
  sprintf(key,"fmri(stop%d)",ev+1); stop=atof(find_line(fn, key, fl));

  skip*=mult; off*=mult; on*=mult; phase*=mult;
  stop*=mult; if ( (stop<0) || (stop+skip>mnpts) ) stop=mnpts-skip;
  
  for(t=skip;t<skip+stop;t++)
    {
      if ( ((int)(t+phase-skip))%((int)(off+on)) >= off )
	X[t*maxcols+Xev]=1.0;
      else
	X[t*maxcols+Xev]=0.0;
    }
}
break;

/* }}} */
    case 1:
      /* {{{ sinusoid */

{
  float skip, period, phase, stop;

  sprintf(key,"fmri(skip%d)",ev+1); skip=atof(find_line(fn, key, fl));
  sprintf(key,"fmri(period%d)",ev+1); period=atof(find_line(fn, key, fl));
  sprintf(key,"fmri(phase%d)",ev+1); phase=atof(find_line(fn, key, fl));
  sprintf(key,"fmri(stop%d)",ev+1); stop=atof(find_line(fn, key, fl));

  skip*=mult; period*=0.5*mult/M_PI; phase*=mult;
  stop*=mult; if ( (stop<0) || (stop+skip>mnpts) ) stop=mnpts-skip;

  for(t=skip;t<skip+stop;t++)
    X[t*maxcols+Xev] = 0.5 * ( 1.0 + sin( M_PI + (t+phase-skip) / period ) );
}
break;

/* }}} */
    case 2:
      /* {{{ custom single column */

{
  FILE *ifp2;

  sprintf(key,"fmri(custom%d)",ev+1); strcpy(filename,(find_line(fn, key, fl)));

  if ((ifp2=fopen(filename,"r"))==NULL)
    {
      printf("Can't open %s for reading\n",filename);
      exit(1);
    }

  for(i=0,t=0; t<npts; t++)
    {
      float tmpf;
	  
      if (fscanf(ifp2,"%f",&tmpf)!=1)
	{
	  printf("Not enough data in %s\n",filename);
	  exit(1);
	}

      for(; i<trmult*(t+1); i++)
	X[i*maxcols+Xev]=tmpf;
    }

  fclose(ifp2);
}
break;

/* }}} */
    case 3:
      /* {{{ custom 3 columns */

{
  FILE *ifp2;
  float start, stop, value;
  int   success=0;

  sprintf(key,"fmri(custom%d)",ev+1); strcpy(filename,(find_line(fn, key, fl)));

  if ((ifp2=fopen(filename,"r"))==NULL)
    {
      printf("Can't open %s for reading\n",filename);
      exit(1);
    }

  while( fscanf(ifp2,"%f %f %f",&start,&stop,&value) == 3 )
    {
      start*=mult; stop=start+stop*mult;

      if ( (stop>0) && (stop-start<1) ) stop=start+1.1; /* if very short stim time make sure it happens */

      if (start<0)     start=0;    /* don't enter stuff before t=0 */
      if (start>mnpts) stop=0;     /* don't do anything */
      if (stop>mnpts)  stop=mnpts; /* don't overrun */

      for(t=start; t<stop; t++)
	{
	  X[t*maxcols+Xev]=value;
	  success=1;
	}
    }

  fclose(ifp2);

  if ( success==0 )
    {
      printf("No valid\n[onset duration strength]\ntriplets found in %s\n",filename);
      exit(1);
    }
}
break;

/* }}} */
    case 4:
      /* {{{ interaction */

{
  int tmp, doit, source_ev, first=1;

  for(t=0;t<npts;t++)
    X[t*maxcols+Xev]=1.0;
  
  for(source_ev=tmp=0; tmp<ev; source_ev++, tmp++)
    {
      sprintf(key,"fmri(interactions%d.%d",ev+1,tmp+1); doit=atoi(find_line(fn, key, fl));
      if (origevs[source_ev]==0) source_ev++;
      if (doit)
	{
	  float themin=X[source_ev];
	  for(t=0;t<npts;t++)
	    if (X[t*maxcols+source_ev]<themin)
	      themin=X[t*maxcols+source_ev];
	  if (first)
	    for(t=0;t<npts;t++)
	      X[t*maxcols+Xev]=X[t*maxcols+source_ev]-themin;
	  else
	    for(t=0;t<npts;t++)
	      X[t*maxcols+Xev]=MIN(X[t*maxcols+source_ev]-themin,X[t*maxcols+Xev]);
	  first=0;
	}
    }
}
break;

/* }}} */
    }

/* }}} */
  if (shape<4)
    {
      /* {{{ convolution */

if (convolve>0)
{
  float *oX=NULL, *cX=NULL;
  int   fw=0;

  /* prepare the convolution kernel */
  switch (convolve)
    {
    case 1:
      /* {{{ Gaussian */

{
  float sigma, delay, tmpf;

  sprintf(key,"fmri(gausssigma%d)",ev+1); sigma=atof(find_line(fn, key, fl));
  sprintf(key,"fmri(gaussdelay%d)",ev+1); delay=atof(find_line(fn, key, fl));

  sigma /= dt;
  delay /= dt;

  fw = delay + sigma*5;
  cX=(float *)malloc(sizeof(float)*fw);

  for(i=0; i<fw; i++)
    {
      tmpf=(((float)i)-delay)/sigma;
      tmpf=exp(-0.5*tmpf*tmpf);
      if (i<1/dt) tmpf *= exp( -pow( 1-0.5*(((float)i)*dt) , 6 ) ); /* between 0 and 1s, window start of kernel */
      cX[i]=tmpf;
    }
}
break;

/* }}} */
    case 2:
      /* {{{ Gamma */

{
  float sigma, delay, alpha, beta;

  sprintf(key,"fmri(gammasigma%d)",ev+1); sigma=atof(find_line(fn, key, fl));
  sprintf(key,"fmri(gammadelay%d)",ev+1); delay=atof(find_line(fn, key, fl));

  sigma /= dt;
  delay /= dt;

  alpha = delay*delay / (sigma*sigma);
  beta  = sigma*sigma / delay;

  fw = delay + sigma*5;
  cX=(float *)malloc(sizeof(float)*fw);

  for(i=0; i<fw; i++)
    {
      cX[i]=gammpdf(alpha,1/beta,(float)i);
      /*      if (i>trmult*2) cX[i]=0;*/
    }
}
break;

/* }}} */
    case 3:
      /* {{{ SPM99 HRF */

{
  float sigma1=2.449/dt, delay1=6/dt, sigma2=4/dt, delay2=16/dt, ratio=6,
    alpha1 = delay1*delay1 / (sigma1*sigma1),
    beta1  = sigma1*sigma1 / delay1,
    alpha2 = delay2*delay2 / (sigma2*sigma2),
    beta2  = sigma2*sigma2 / delay2;

  fw = delay2 + sigma2*5;
  cX=(float *)malloc(sizeof(float)*fw);

  for(i=0; i<fw; i++)
    cX[i]= gammpdf(alpha1,1/beta1,(float)i) - gammpdf(alpha2,1/beta2,(float)i)/ratio;
}
break;

/* }}} */
    }

  /* {{{ renorm kernel */

{
  double kernelsum=0;

  for(i=0; i<fw; i++)
    kernelsum += cX[i]; /* note that in general maybe this should be fabs(cX[i]), but looking at SPM99HRF output, I've not done this */

  for(i=0; i<fw; i++)
    cX[i] /= kernelsum;
}

/* }}} */
  /* {{{ carry out the convolution */

  oX=(float *)malloc(sizeof(float)*(mnpts+fw));

  for(t=0;t<fw;t++) /* pad backwards with zeroes */
    oX[t]=0; /*X[maxcols+Xev];*/
  for(t=fw;t<mnpts+fw;t++)
    oX[t]=X[(t-fw)*maxcols+Xev];

  for(t=0;t<mnpts;t++)
    {
      float sumf=0;
      for(i=0; i<fw; i++)
	sumf += oX[t+fw-i] * cX[i];
      X[t*maxcols+Xev]=sumf;
    }

/* }}} */
}

/* }}} */
      /* {{{ COMMENT square the thingy */

#ifdef FoldingComment

  for(t=0;t<mnpts;t++)
    {
      X[t*maxcols+Xev]=2*X[t*maxcols+Xev]-X[t*maxcols+Xev]*X[t*maxcols+Xev];
    }

#endif

/* }}} */
      /* {{{ resample down in time */

/* sample in the middle of the upsampled period */

for(t=0;t<npts;t++)
     X[t*maxcols+Xev] = X[((int)((t+0.5)*trmult))*maxcols+Xev];

/* }}} */
    }
  /* {{{ temporal filtering */

/* {{{ COMMENT bugger up data to check temporal filtering */

#ifdef FoldingComment

for(ev=0; ev<Xev; ev++)
  for(t=0;t<npts;t++)
     X[t*maxcols+ev] += 0.0002*((float)t-npts/2)*((float)t-npts/2);

#endif

/* }}} */

if ( ( (templp_yn) || (temphp_yn) ) && (tempfiltyn) )
{
  image_struct im;
  double hp_sigma=-1, lp_sigma=-1;

  im.x=im.y=im.z=1;
  im.t=npts;
  im.i=(float*)malloc(sizeof(float)*npts);

  if (templp_yn) lp_sigma=2.8/tr;
  if (temphp_yn) hp_sigma=0.5*nltffwhm/tr;
  /*printf("%f %f\n",hp_sigma,lp_sigma);*/

  for(t=0;t<npts;t++)
    im.i[t] = X[t*maxcols+Xev];
  tempfilt(im, hp_sigma, lp_sigma);
  for(t=0;t<npts;t++)
    X[t*maxcols+Xev] = im.i[t];
}

/* }}} */
  /* {{{ orthogonalisation */

{
  int tmp, source_ev, orthsource, northsource, *ortho=malloc(sizeof(int)*ev);
  double *W=calloc((npts+maxcols)*maxcols,sizeof(double)),
    *Z=calloc(maxcols,sizeof(double));

  /* {{{ find out which EVs we will orthogonalise WRT, and count them */

  for(tmp=northsource=0; tmp<ev; tmp++)
    {
      sprintf(key,"fmri(ortho%d.%d)",ev+1,tmp+1); ortho[tmp]=atoi(find_line(fn, key, fl));
      northsource+=ortho[tmp];
    }

/* }}} */
  /* {{{ copy these EVs into the SVD input array */

  for(source_ev=tmp=orthsource=0; tmp<ev; source_ev++, tmp++)
    {
      if (origevs[source_ev]==0) source_ev++;
      if (ortho[tmp])
	{
	  for(t=0;t<npts;t++)
	    W[t*northsource+orthsource]=X[t*maxcols+source_ev];
	  orthsource++;
	}
    }

/* }}} */
  /* {{{ run SVD */

if (northsource>1)
     SVD(W,Z,npts,northsource);

/* {{{ COMMENT replace the orthogonalising source EVs with the SVD output, for debugging */

#ifdef FoldingComment

  for(source_ev=tmp=orthsource=0; tmp<ev; source_ev++, tmp++)
    {
      if (origevs[source_ev]==0) source_ev++;
      if (ortho[tmp])
	{
	  for(t=0;t<npts;t++)
	    X[t*maxcols+source_ev]=W[t*northsource+orthsource];
	  orthsource++;
	}
    }

#endif

/* }}} */

/* }}} */
  /* {{{ orthogonalise current EV WRT SVD output */

for(tmp=0; tmp<northsource; tmp++)
{
  float AdotA=0, BdotA=0;
  
  for(t=0;t<npts;t++)
    {
      AdotA += W[t*northsource+tmp]*W[t*northsource+tmp];
      BdotA += X[t*maxcols+Xev]*W[t*northsource+tmp];
    }
  
  for(t=0;t<npts;t++)
    X[t*maxcols+Xev] -= BdotA*W[t*northsource+tmp]/AdotA;
}

/* }}} */
}

/* }}} */
  /* {{{ demean */

{
  double Xmean=0;
  
  for(t=0;t<npts;t++)
    Xmean += X[t*maxcols+Xev];
  
  Xmean /= npts;
  
  for(t=0;t<npts;t++)
    X[t*maxcols+Xev] -= Xmean;
}

/* }}} */
  if (deriv_yn) Xev++;
}

/* {{{ produce temporal derivs */

for(ev=0; ev<Xev; ev++)
  if (origevs[ev]==0)
  {
    /* do end points differently */
    X[ev] =                  X[maxcols+ev-1]          - X[ev-1];
    X[(npts-1)*maxcols+ev] = X[(npts-1)*maxcols+ev-1] - X[(npts-2)*maxcols+ev-1];

    for(t=1;t<npts-1;t++)
      X[t*maxcols+ev] = (X[(t+1)*maxcols+ev-1] - X[(t-1)*maxcols+ev-1])*0.5;
  }

/* }}} */
/* {{{ demean */

for(ev=0; ev<Xev; ev++)
  {
    double Xmean=0;

    for(t=0;t<npts;t++)
	Xmean += X[t*maxcols+ev];

    Xmean /= npts;

    for(t=0;t<npts;t++)
      X[t*maxcols+ev] -= Xmean;
  }

/* }}} */
/* {{{ add final (constant) column */

if (constcol)
{
  for(t=0;t<npts;t++)
    X[t*maxcols+Xev]=1.0;

  origevs[Xev]=0;
  Xev++;
}

/* }}} */

/* }}} */
  /* {{{ read contrasts */

C=(float *)malloc(sizeof(float)*Xev*ncon);

for(con=0; con<ncon; con++)
  for(tmp=ev=0; ev<Xev; ev++)
  {
    if (origevs[ev]==1)
      {
	sprintf(key,"fmri(con%d.%d)",con+1,tmp+1); C[con*Xev+ev]=atof(find_line(fn, key, fl));
	tmp++;
      }
    else
      C[con*Xev+ev]=0;
  }

/* }}} */
  /* {{{ read ftests */

F=malloc(sizeof(int)*ncon*nftests);

for(f=0; f<nftests; f++)
  for(con=0; con<ncon; con++)
  {
    sprintf(key,"fmri(ftest%d.%d)",f+1,con+1); F[f*ncon+con]=atoi(find_line(fn, key, fl));
  }

/* }}} */
  /* {{{ write matrix */

sprintf(filename,"%s.mat",argv[1]);
if ((ofp=fopen(filename,"w"))==NULL)
{
  printf("Can't open %s for writing\n",filename);
  exit(1);
}

fprintf(ofp,"%%! VEST-Waveform File\n");
fprintf(ofp,"/NumWaves	%d\n",Xev);
fprintf(ofp,"/NumPoints	%d\n",npts);
fprintf(ofp,"/Skip		0\n");

for(ev=0; ev<Xev; ev++)
  fprintf(ofp,"/Custom	0 %s\n",argv[1]); /* argv[1] is actually a dummy entry - the values are really used from this file itself */

fprintf(ofp,"\n/Matrix\n");

for(t=0; t<npts; t++)
{
  for(ev=0; ev<Xev; ev++)
    fprintf(ofp,"%e	",X[t*maxcols+ev]);
  fprintf(ofp,"\n");
}

fclose(ofp);

/* }}} */
  /* {{{ write contrasts */

sprintf(filename,"%s.con",argv[1]);
if ((ofp=fopen(filename,"w"))==NULL)
{
  printf("Can't open %s for writing\n",filename);
  exit(1);
}

fprintf(ofp,"%%! VEST-Waveform File\n");
fprintf(ofp,"/NumWaves	%d\n",Xev);
fprintf(ofp,"/NumContrasts	%d\n",ncon);
fprintf(ofp,"\n/Matrix\n");

for(con=0; con<ncon; con++)
{
  for(ev=0; ev<Xev; ev++)
    fprintf(ofp,"%e ",C[con*Xev+ev]);
  fprintf(ofp,"\n");
}

fclose(ofp);

/* }}} */
  /* {{{ write ftests */

if (nftests>0)
{
  sprintf(filename,"%s.fts",argv[1]);
  if ((ofp=fopen(filename,"w"))==NULL)
    {
      printf("Can't open %s for writing\n",filename);
      exit(1);
    }

  fprintf(ofp,"%%! VEST-Waveform File\n");
  fprintf(ofp,"/NumWaves	%d\n",ncon);
  fprintf(ofp,"/NumContrasts	%d\n",nftests);
  fprintf(ofp,"\n/Matrix\n");

  for(f=0; f<nftests; f++)
    {
      for(con=0; con<ncon; con++)
	fprintf(ofp,"%d ",F[f*ncon+con]);
      fprintf(ofp,"\n");
    }

  fclose(ofp);
}

/* }}} */
  /* {{{ svd to check rank of matrix */

if (Xev>1)
{
  W=(double *)malloc(sizeof(double)*(npts+Xev)*Xev);
  Z=(double *)calloc(Xev,sizeof(double));
    
  for(ev=0; ev<Xev; ev++)
    for(t=0;t<npts;t++)
      W[t*Xev+ev]=X[t*maxcols+ev];

  SVD(W,Z,npts,Xev);
  for(ev=0; ev<Xev; ev++)
    Z[ev]=sqrt(Z[ev]);
  minl=maxl=Z[0];
  for(ev=0; ev<Xev; ev++)
    {
      /*printf("%e\n",Z[ev]);*/
      if (Z[ev]<minl) minl=Z[ev];
      if (Z[ev]>maxl) maxl=Z[ev];
    }
  condition=minl/maxl;

  /*printf("condition = %e\n",condition);*/

  if (condition<MINCONDITION)
    printf("At least one EV is (close to) a linear combination of the others. You should probably alter your design.\n\n(Design matrix is rank deficient - ratio of min:max eigenvalues in SVD of matrix is %e)\n",condition);
}

/* }}} */
  /* {{{ write autocov matrix */

if (Xev>1)
{
  /* {{{ setup vars */

  FILE *ofp2;
  unsigned char *r,*g,*b;
  int border=5, mag, size, xsize, ysize, x, y, evx, evy;
  float *cov=malloc(Xev*Xev*sizeof(float)), *covnorm=malloc(Xev*sizeof(float));

  mag  = MIN( MAX( 1 , 300/Xev  ) , 50);

  size  = Xev*mag;
  xsize = size*2 + border*3;
  ysize = size   + border*2;

/* }}} */
  /* {{{ malloc images and fill in background */

r=(unsigned char *)malloc(xsize*ysize);
g=(unsigned char *)malloc(xsize*ysize);
b=(unsigned char *)malloc(xsize*ysize);

memset((void *)r,(unsigned char)180,xsize*ysize);
memset((void *)g,(unsigned char)215,xsize*ysize);
memset((void *)b,(unsigned char)255,xsize*ysize);

/* }}} */
  /* {{{ cov matrix orig */

for(evx=0; evx<Xev; evx++)
  for(evy=0; evy<Xev; evy++)
  {
    cov[evy*Xev+evx]=0;
    for(t=0;t<npts;t++)
      cov[evy*Xev+evx]+=X[t*maxcols+evx]*X[t*maxcols+evy];
    if (evx==evy)
      covnorm[evx]=cov[evx*Xev+evx];
  }

for(evx=0; evx<Xev; evx++)
  for(evy=0; evy<Xev; evy++)
  {
    tmpf=covnorm[evy]*covnorm[evx];
    if (tmpf<=0)
      cov[evy*Xev+evx]=0;
    else
      cov[evy*Xev+evx]=ABS(cov[evy*Xev+evx])/sqrt(tmpf);
  }

for(evx=0; evx<Xev; evx++)
  for(evy=0; evy<Xev; evy++)
    for(y=0; y<mag; y++)
      for(x=0; x<mag; x++)
        r[(border+evy*mag+y)*xsize+border+evx*mag+x] =
        g[(border+evy*mag+y)*xsize+border+evx*mag+x] =
        b[(border+evy*mag+y)*xsize+border+evx*mag+x] = 255*cov[evy*Xev+evx];

/* }}} */
  /* {{{ cov matrix SVD */

tmpf=0;
for(evx=0; evx<Xev; evx++)
  for(evy=0; evy<Xev; evy++)
  {
    if (evx!=evy)
      cov[evy*Xev+evx]=0;
    else
      {
	cov[evy*Xev+evx]=Z[evx];
	if (cov[evy*Xev+evx]>tmpf)
	  tmpf=cov[evy*Xev+evx];
      }
  }

for(evx=0; evx<Xev; evx++)
  cov[evx*Xev+evx]/=tmpf;

for(evx=0; evx<Xev; evx++)
  for(evy=0; evy<Xev; evy++)
    for(y=0; y<mag; y++)
      for(x=0; x<mag; x++)
        r[(border+evy*mag+y)*xsize+2*border+size+evx*mag+x] =
        g[(border+evy*mag+y)*xsize+2*border+size+evx*mag+x] =
        b[(border+evy*mag+y)*xsize+2*border+size+evx*mag+x] = 255*cov[evy*Xev+evx];

/* }}} */
  /* {{{ output image */

sprintf(filename,"%s_cov.ppm",argv[1]);
if ((ofp2=fopen(filename,"w"))==NULL)
{
  printf("Can't open %s for writing\n",filename);
  exit(1);
}

fprintf(ofp2,"P6\n");
fprintf(ofp2,"%d %d\n",xsize,ysize);
fprintf(ofp2,"255\n");

for(y=0; y<ysize; y++)
     for(x=0; x<xsize; x++)
{
  fwrite(&r[y*xsize+x],1,1,ofp2);
  fwrite(&g[y*xsize+x],1,1,ofp2);
  fwrite(&b[y*xsize+x],1,1,ofp2);
}

fclose(ofp2);

/* }}} */
}

/* }}} */
  /* {{{ write image preview (note - corrupts X[]) */

{
  /* {{{ setup vars */

  FILE *ofp2;
  unsigned char *r,*g,*b;
  int xmag, ymag, ymag2, border=5, temp=9, fsize=9, xsize, ysize, xx, yy, x, y;

  xmag  = MIN( MAX( 1 , 600/Xev  ) , 75);
  ymag  = MIN( MAX( 1 , 400/npts ) , 20);
  ymag2 = MIN( MAX( 1 , 150/ncon ) , 30);

  xsize=Xev*xmag + border*(Xev+3+nftests+(nftests>0)) + temp + nftests*fsize;
  ysize=npts*ymag + ncon*ymag2 + border*(3+ncon);

/* }}} */
  /* {{{ reset X[] range (but don't change offset) */

for(ev=0; ev<Xev; ev++)
{
  float absmax=0;

  for(t=0;t<npts;t++)
    if ( fabs(X[t*maxcols+ev]) > absmax )
      absmax=fabs(X[t*maxcols+ev]);

  if (absmax>0)
    for(t=0;t<npts;t++)
      X[t*maxcols+ev] /= absmax;
}

/* }}} */
  /* {{{ malloc images and fill in background */

r=(unsigned char *)malloc(xsize*ysize);
g=(unsigned char *)malloc(xsize*ysize);
b=(unsigned char *)malloc(xsize*ysize);

memset((void *)r,(unsigned char)180,xsize*ysize);
memset((void *)g,(unsigned char)215,xsize*ysize);
memset((void *)b,(unsigned char)255,xsize*ysize);

/* }}} */
  /* {{{ time */

for(yy=0; yy<npts; yy++)
  for(y=0; y<ymag; y++)
    for(x=0; x<temp; x++)
    {
      int intensity=64, rintensity;

      if (yy%10==0) intensity=255;

      rintensity=intensity;
      if (abs(yy-npts/2)<=(0.5*nltffwhm/tr))
	{
	  if (x>1&&x<7)
	    rintensity=intensity=0;
	  if (x>2&&x<6)
	    {
	      rintensity=255;
	      intensity=0;
	    }
	}

      r[(border+yy*ymag+y)*xsize+border+x]=rintensity;
      g[(border+yy*ymag+y)*xsize+border+x]=intensity;
      b[(border+yy*ymag+y)*xsize+border+x]=intensity;
    }

/* }}} */
  /* {{{ ftests */

for(f=0; f<nftests; f++)
  for(con=0; con<ncon; con++)
    {
      for(y=0; y<fsize; y++)
        for(x=0; x<fsize; x++)
          r[(int)( ((3+con)*border+(con+0.5)*ymag2-fsize/2+npts*ymag+y))*xsize +
		   border*(Xev+4)+Xev*xmag+x+temp+f*(border+fsize) ]=
	  g[(int)( ((3+con)*border+(con+0.5)*ymag2-fsize/2+npts*ymag+y))*xsize +
		   border*(Xev+4)+Xev*xmag+x+temp+f*(border+fsize) ]=
  	  b[(int)( ((3+con)*border+(con+0.5)*ymag2-fsize/2+npts*ymag+y))*xsize +
		   border*(Xev+4)+Xev*xmag+x+temp+f*(border+fsize) ]=0;

      if (F[f*ncon+con])
	for(y=1; y<fsize-1; y++)
	  for(x=1; x<fsize-1; x++)
	    r[(int)( ((3+con)*border+(con+0.5)*ymag2-fsize/2+npts*ymag+y))*xsize +
		     border*(Xev+4)+Xev*xmag+x+temp+f*(border+fsize) ]=255;
    }

/* }}} */
  /* {{{ grey */

  if (grey_yn)
    {
      /* {{{ DM: grey */

  for(yy=0; yy<npts; yy++)
    for(xx=0; xx<Xev; xx++)
      for(y=0; y<ymag; y++)
	for(x=0; x<xmag; x++)
	  {
	    float XX = X[yy*maxcols+xx];

	    r[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+temp]=(XX * 100.0)+128;
	    g[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+temp]=(XX * 100.0)+128;
	    b[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+temp]=(XX * 100.0)+128;
	  }

/* }}} */
      /* {{{ Con: grey */

  for(yy=0; yy<ncon; yy++)
    for(xx=0; xx<Xev; xx++)
      for(y=0; y<ymag2; y++)
	for(x=0; x<xmag; x++)
	  {
	    float XX = C[yy*Xev+xx];

	    if (XX<-1) XX=-1;
	    if (XX> 1) XX=1;

	    r[((3+yy)*border+yy*ymag2+npts*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+temp]=(XX * 100.0)+128;
	    g[((3+yy)*border+yy*ymag2+npts*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+temp]=(XX * 100.0)+128;
	    b[((3+yy)*border+yy*ymag2+npts*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+temp]=(XX * 100.0)+128;
	  }

/* }}} */
    }
  else
    {
      /* {{{ DM: not grey */

  for(yy=0; yy<npts; yy++)
    for(xx=0; xx<Xev; xx++)
      for(y=0; y<ymag; y++)
	for(x=0; x<xmag; x++)
	  {
	    r[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+temp]=0;
	    g[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+temp]=0;
	    b[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+temp]=0;
	  }

/* }}} */
      /* {{{ Con: not grey */

  for(yy=0; yy<ncon; yy++)
    for(xx=0; xx<Xev; xx++)
      for(y=0; y<ymag2; y++)
	for(x=0; x<xmag; x++)
	  {
	    r[((3+yy)*border+yy*ymag2+npts*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+temp]=0;
	    g[((3+yy)*border+yy*ymag2+npts*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+temp]=0;
	    b[((3+yy)*border+yy*ymag2+npts*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+temp]=0;
	  }

/* }}} */
    }

/* }}} */
  /* {{{ red */

  if (red_yn)
    {
      /* {{{ DM: red */

  for(yy=0; yy<npts; yy++)
    for(xx=0; xx<Xev; xx++)
      for(y=0; y<ymag; y++)
	{
	  float XX = X[yy*maxcols+xx];
	  
	  x = (XX*0.8+1.0)*xmag/2;

	  r[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+temp]=255;
	  g[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+temp]=0;
	  b[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+temp]=0;
	  r[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+1+temp]=255;
	  g[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+1+temp]=0;
	  b[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+1+temp]=0;
	  r[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x-1+temp]=255;
	  g[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x-1+temp]=0;
	  b[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x-1+temp]=0;
	  r[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+2+temp]=0;
	  g[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+2+temp]=0;
	  b[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+2+temp]=0;
	  r[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x-2+temp]=0;
	  g[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x-2+temp]=0;
	  b[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x-2+temp]=0;
	}

/* }}} */
      /* {{{ Con: red */

  for(yy=0; yy<ncon; yy++)
    for(xx=0; xx<Xev; xx++)
      for(x=0; x<xmag; x++)
        {
	  float XX = C[yy*Xev+xx];

	  if (XX<-1) XX=-1;
	  if (XX> 1) XX=1;
	    
	  y = (-XX*0.9+1.0)*ymag2/2;

	  r[((3+yy)*border+yy*ymag2+npts*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+temp]=255;
	  g[((3+yy)*border+yy*ymag2+npts*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+temp]=0;
	  b[((3+yy)*border+yy*ymag2+npts*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+temp]=0;
	  r[((3+yy)*border+yy*ymag2+npts*ymag+y+1)*xsize+border*(xx+3)+xx*xmag+x+temp]=255;
	  g[((3+yy)*border+yy*ymag2+npts*ymag+y+1)*xsize+border*(xx+3)+xx*xmag+x+temp]=0;
	  b[((3+yy)*border+yy*ymag2+npts*ymag+y+1)*xsize+border*(xx+3)+xx*xmag+x+temp]=0;
	  r[((3+yy)*border+yy*ymag2+npts*ymag+y-1)*xsize+border*(xx+3)+xx*xmag+x+temp]=255;
	  g[((3+yy)*border+yy*ymag2+npts*ymag+y-1)*xsize+border*(xx+3)+xx*xmag+x+temp]=0;
	  b[((3+yy)*border+yy*ymag2+npts*ymag+y-1)*xsize+border*(xx+3)+xx*xmag+x+temp]=0;
	  r[((3+yy)*border+yy*ymag2+npts*ymag+y+2)*xsize+border*(xx+3)+xx*xmag+x+temp]=0;
	  g[((3+yy)*border+yy*ymag2+npts*ymag+y+2)*xsize+border*(xx+3)+xx*xmag+x+temp]=0;
	  b[((3+yy)*border+yy*ymag2+npts*ymag+y+2)*xsize+border*(xx+3)+xx*xmag+x+temp]=0;
	  r[((3+yy)*border+yy*ymag2+npts*ymag+y-2)*xsize+border*(xx+3)+xx*xmag+x+temp]=0;
	  g[((3+yy)*border+yy*ymag2+npts*ymag+y-2)*xsize+border*(xx+3)+xx*xmag+x+temp]=0;
	  b[((3+yy)*border+yy*ymag2+npts*ymag+y-2)*xsize+border*(xx+3)+xx*xmag+x+temp]=0;
	}

/* }}} */
    }

/* }}} */
  /* {{{ output image */

sprintf(filename,"%s.ppm",argv[1]);
if ((ofp2=fopen(filename,"w"))==NULL)
{
  printf("Can't open %s for writing\n",filename);
  exit(1);
}

fprintf(ofp2,"P6\n");
fprintf(ofp2,"%d %d\n",xsize,ysize);
fprintf(ofp2,"255\n");

for(y=0; y<ysize; y++)
     for(x=0; x<xsize; x++)
{
  fwrite(&r[y*xsize+x],1,1,ofp2);
  fwrite(&g[y*xsize+x],1,1,ofp2);
  fwrite(&b[y*xsize+x],1,1,ofp2);
}

fclose(ofp2);

/* }}} */
}

/* }}} */

  return(0);
}

/* }}} */
