/* {{{ Copyright etc. */

/*  tsplot - FMRI time series and model plotting

    Stephen Smith and Mark Woolrich, FMRIB Image Analysis Group

    Copyright (C) 1999-2001 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

/* }}} */
/* {{{ defines, includes and typedefs */

#include "featlib.h"

/* }}} */
/* {{{ find_key */

char *find_key(FILE *fd, char *control_line, const char *key)
{
  while (fgets(control_line, 1000, fd)!=NULL)
    if ( (strncmp(control_line,key,strlen(key))==0) )
      {
	return control_line+strlen(key);
      }
	  
  printf("Error: key \"%s\" not found.\n",key);
  exit(1);
}

/* }}} */
/* {{{ establish_pwfilter */

void establish_pwfilter(const ColumnVector& ac, ColumnVector& pwfilter, int zeropad, int npts)
{
  // FFT auto corr estimate
  ColumnVector dummy(zeropad);
  dummy = 0;
  ColumnVector vrow(zeropad);
  vrow = 0;

  // ac maybe cutoff to be smaller than npts
  int sizeTS = ac.Nrows();
  if (sizeTS > npts/2)
    sizeTS = npts/2;
  
  vrow.Rows(1,sizeTS) = ac.Rows(1,sizeTS);
  vrow.Rows(zeropad - sizeTS + 2, zeropad) = ac.Rows(2, sizeTS).Reverse();

  ColumnVector ac_fft_imag;
    
  FFT(vrow, dummy, pwfilter, ac_fft_imag);      

  // inverse auto corr to give prewhitening filter
  // no DC component so set first value to 0
  pwfilter(1) = 0.0;
  
  for(int j = 2; j <= zeropad; j++)
    {
      if (pwfilter(j)<0)
	{
	  cout << "Warning: possible high autocorrelation in time series" << endl;
	  pwfilter(j)=0;
	}
      else
	pwfilter(j) = 1.0/sqrt(pwfilter(j));	      
    }  

  // normalise pwfilter such that sum(j)((pwfilter)^2/zeropad)) = 1
  pwfilter /= sqrt(pwfilter.SumSquare()/zeropad);
}

/* }}} */
/* {{{ prewhiten_data */

void prewhiten_data(const ColumnVector& data, ColumnVector& pwdata, ColumnVector& pwfilter, int zeropad, int npts)
{ 
  ColumnVector data_fft_real, data_fft_imag, realifft, dummy;
  dummy.ReSize(zeropad);
  dummy = 0;

  // Remove and store mean
  float mn = MISCMATHS::mean(data);
  pwdata.ReSize(zeropad);
  pwdata = 0;
  pwdata.Rows(1,npts) = data - mn;

  // FFT data
  FFT(pwdata, dummy, data_fft_real, data_fft_imag);
  FFTI(SP(pwfilter, data_fft_real), SP(pwfilter, data_fft_imag), realifft, dummy); 
  // take first npts and restore mean
  pwdata = realifft.Rows(1,npts) + mn;  
}

/* }}} */
/* {{{ prewhiten_model */

void prewhiten_model(const ColumnVector& ac, double *model, double *pwmodel, int nevs, int npts)
{
  Matrix dm(npts, nevs), pwdm(npts, nevs);

  // Get values out of model
  for(int ev=0; ev<nevs; ev++)
    {
      for(int t=0; t<npts; t++)
	{
	  dm(t+1,ev+1) = model[t*nevs+ev];
	}
    }

  pwdm = dm;

  int zeropad = (int)pow(2,ceil(log(npts)/log(2)));
  
  // get prewhitening filter from ac
  ColumnVector pwfilter;
  establish_pwfilter(ac, pwfilter, zeropad, npts);

  // prewhiten each of the evs
  for(int ev = 1; ev <= nevs; ev++)
    {
      ColumnVector pwdata;
      prewhiten_data(dm.Column(ev), pwdata, pwfilter, zeropad, npts);
      pwdm.Column(ev) = pwdata;
    }

  // setup and return result
  for(int ev=0; ev<nevs; ev++)
    {
      for(int t=0; t<npts; t++)
	{
	  pwmodel[t*nevs+ev] = pwdm(t+1,ev+1);
	}
    }
}

/* }}} */
/* {{{ prewhiten_timeseries */

void prewhiten_timeseries(const ColumnVector& ac, const ColumnVector& ts, ColumnVector& pwts, int npts)
{
  int zeropad = (int)pow(2,ceil(log(npts)/log(2)));
  
  // get prewhitening filter from ac
  ColumnVector pwfilter;
  establish_pwfilter(ac, pwfilter, zeropad, npts);

  // prewhiten
  prewhiten_data(ts, pwts, pwfilter, zeropad, npts);
}

/* }}} */
/* {{{ read_model */

double *read_model(char *filename, int *nevs, int *npts)
{
  FILE *designfp;
  char control_line[1010];
  double *model;
  int i;

  if((designfp=fopen(filename,"r"))==NULL)
    {
      *npts=0;
      return NULL;
    }

  *nevs=(int)atof(find_key(designfp,control_line,"/NumWaves"));
  *npts=(int)atof(find_key(designfp,control_line,"/NumPoints"));

  atof(find_key(designfp,control_line,"/Matrix"));

  model=(double*)malloc(sizeof(double)* *nevs * *npts);

  for(i=0; i<*nevs * *npts; i++)
    fscanf(designfp,"%lf",&model[i]);

  return model;
}

/* }}} */
/* {{{ read_contrasts */

double *read_contrasts(char *filename, int *nevs, int *ncon)
{
  FILE *designfp;
  char control_line[1010];
  double *contrasts;
  int i;

  if((designfp=fopen(filename,"r"))==NULL)
    {
      fprintf(stderr,"Can't open %s\n",filename);
      exit(1);
    }

  *nevs=(int)atof(find_key(designfp,control_line,"/NumWaves"));
  *ncon=(int)atof(find_key(designfp,control_line,"/NumContrasts"));

  atof(find_key(designfp,control_line,"/Matrix"));

  contrasts=(double*)malloc(sizeof(double)* *nevs * *ncon);

  for(i=0; i<*nevs * *ncon; i++)
    fscanf(designfp,"%lf",&contrasts[i]);

  return contrasts;
}

/* }}} */
/* {{{ read_ftests */

void read_ftests(char *filename, int *nftests)
{
  FILE *designfp;
  char control_line[1010];

  *nftests=0;

  if((designfp=fopen(filename,"r"))!=NULL)
    *nftests=(int)atof(find_key(designfp,control_line,"/NumContrasts"));
}

/* }}} */
