/* {{{ Copyright etc. */

/*  tsplot - FMRI time series and model plotting

    Stephen Smith and Mark Woolrich, FMRIB Image Analysis Group

    Copyright (C) 1999-2001 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

/* }}} */
/* {{{ defines, includes and typedefs */

#include "featlib.h"

/* }}} */
/* {{{ usage */

void usage(void)
{
  printf("Usage: tsplot <feat_directory.feat> [options]\n");
  printf("[-f <4D_data>] input main filtered data, in case it's not in the FEAT directory\n");
  printf("[-c <X Y Z>] : use X,Y,Z instead of max Z stat position\n");
  printf("[-C <X Y Z output_file.txt>] : use X,Y,Z to output time series only - no stats or modelling\n");
  printf("[-o <output_directory>] change output directory from default of input feat directory\n");
  printf("[-n] don't weight cluster averaging with Z stats\n");
  exit(1);
}

/* }}} */
/* {{{ main */

int main(int argc, char **argv)
{
  /* {{{ variables */

FILE         *ifp, *ofp;
int          argi=1, prewhiten=1, ev, i, j, t, X=0, Y=0, Z=0, nevs, npts,
  ncon=1, nftests=0, size, ymin, ymax, coordset=0, dataonly=0, modelfree=0, zweight=1;
double       *model, *contrasts=NULL, tsmean, tmpf, maxz;
char         fmridata[10000], featdir[10000], outputdir[10000], gpstart[10000], gpname[10000],
             thestring[10000], thestring2[10000], thestring3[10000], datafile[10000], binstr[10000], statname[10000];
ColumnVector pwts;

/* }}} */

  /* {{{ process arguments */

if (argc<2) usage();

strcpy(featdir,argv[argi++]);

strcpy(outputdir,featdir);

sprintf(fmridata,"%s/filtered_func_data",featdir);

/* will use binstr to work out where "gnuplot" and "convert" are, for later system calls */
sprintf(binstr,"%s",argv[0]);
binstr[strlen(binstr)-strlen("tsplot")]=0;

for (;argi<argc;argi++)
{
  if (!strcmp(argv[argi], "-f"))
    /* {{{ alternative fmri data */

{
  argi++;
  if (argc<argi+1)
    {
      printf("Error: no value given following -f\n");
      usage();
    }
  strcpy(fmridata,argv[argi]);
}

/* }}} */
  else if (!strcmp(argv[argi], "-c"))
    /* {{{ alternative voxel position */

{
  coordset=1;
  argi++;
  if (argc<argi+3) /* options following c haven't been given */
    {
      printf("Error: incomplete values given following -c\n");
      usage();
    }
  X=atoi(argv[argi++]);
  Y=atoi(argv[argi++]);
  Z=atoi(argv[argi]);
}

/* }}} */
  else if (!strcmp(argv[argi], "-C"))
    /* {{{ output data only */

{
  coordset=1;
  dataonly=1;
  argi++;
  if (argc<argi+4)
    {
      printf("Error: incomplete values given following -C\n");
      usage();
    }
  X=atoi(argv[argi++]);
  Y=atoi(argv[argi++]);
  Z=atoi(argv[argi++]);
  strcpy(datafile,argv[argi]);
}

/* }}} */
  else if (!strcmp(argv[argi], "-o"))
    /* {{{ output dir */

{
  argi++;
  if (argc<argi+1)
    {
      printf("Error: no value given following -o\n");
      usage();
    }
  strcpy(outputdir,argv[argi]);
}

/* }}} */
  else if (!strcmp(argv[argi], "-n"))
    /* {{{ zweight clusters? */

{
  zweight=0;
}

/* }}} */
}

/* }}} */
  /* {{{ read filtered_func_data */

volume4D<int> im;
read_volume4D(im, fmridata);

/*printf("FMRI dimensions %d %d %d %d\n",image.x,image.y,image.z,image.t);*/

size=im.nvoxels();

if (dataonly)
  /* {{{ output raw data and exit */

{
  if((ofp=fopen(datafile,"w"))==NULL)
    {
      fprintf(stderr,"Can't open output data file %s\n",datafile);
      exit(1);
    }

  for(t=0; t<im.tsize(); t++)
    fprintf(ofp,"%e\n",(float)im(X,Y,Z,t));
  //[t*size+Z*image.x*image.y+Y*image.x+X]);

  fclose(ofp);

  return 0;
}

/* }}} */

/* }}} */
  /* {{{ read design.mat */

sprintf(thestring,"%s/design.mat",featdir);
model=read_model(thestring,&nevs,&npts);

if (npts==0) modelfree=1;

npts=im.tsize();

/* }}} */
  /* {{{ read auto correlation estimates for prewhitening */

double* pwmodel=0;
volume4D<float> acs;
sprintf(thestring,"%s/stats/threshac1.hdr",featdir);
if((ifp=fopen(thestring,"r"))!=NULL)
{
  fclose(ifp);
  sprintf(thestring,"%s/stats/threshac1",featdir);
  read_volume4D(acs, thestring);
  // allocate prewhiten variable for later use
  pwmodel=new double[nevs*npts];
}
else
{ /* no autocorr estimates - thus will use the uncorrected design matrix */
  prewhiten=0;
}

/* }}} */
  /* {{{ read design.con and PEs */

vector< volume<float> > impe(nevs);
if (!modelfree)
{
  sprintf(thestring,"%s/design.con",featdir);
  contrasts=read_contrasts(thestring,&nevs,&ncon);

  for(i=1;i<=nevs;i++)
    {
      sprintf(thestring,"%s/stats/pe%d",featdir,i);
      read_volume(impe[i-1], thestring);
    }
}

/* }}} */
  /* {{{ read design.fts */

if (!modelfree)
{
  sprintf(thestring,"%s/design.fts",featdir);
  read_ftests(thestring,&nftests);
}

/* }}} */
  /* {{{ create plot(s) for each contrast */

for(j=0;j<2;j++)
{
  /* {{{ setup stats type */

int maxi;

if (j==0) { sprintf(statname,"zstat");  maxi=ncon; }
else      { sprintf(statname,"zfstat"); maxi=nftests; }

/* }}} */

  for(i=0; i<maxi; i++)
    {
      /* {{{ vars */

      double *data=(double*)calloc(npts,sizeof(double));
      
      volume<float> imcope;
      volume<float> imz, imw;
      volume<float> imclus;
      
      bool haveclusters=false;

/* }}} */
      /* {{{ read COPE and derived stats; test for f-test output */

{
  FILE *ifp;

  sprintf(thestring,"%s/stats/%s%d.hdr",featdir,statname,i+1);
  if((ifp=fopen(thestring,"r"))!=NULL)
    {
      fclose(ifp);
      sprintf(thestring,"%s/stats/%s%d",featdir,statname,i+1);
      read_volume(imz,thestring);
      imw=imz;
      if (!zweight)
	imw=1;
    }
  else
    continue; /* f-test i wasn't valid - no zfstat image */

  if ( (j==0) && (!modelfree) )
    {
      sprintf(thestring,"%s/stats/cope%d",featdir,i+1);
      read_volume(imcope,thestring);
    }

  if (!coordset)
    {
      sprintf(thestring,"%s/cluster_mask_%s%d.hdr",featdir,statname,i+1);
      if((ifp=fopen(thestring,"r"))!=NULL)
	{
	  fclose(ifp);
	  sprintf(thestring,"%s/cluster_mask_%s%d",featdir,statname,i+1);
	  read_volume(imclus,thestring);
	  haveclusters=(imclus.max()>0);
	}
    }
}

/* }}} */
      /* {{{ open data file */

sprintf(thestring,"%s/tsplot_%s%d.txt",outputdir,statname,i+1);

if((ofp=fopen(thestring,"w"))==NULL)
{
  fprintf(stderr,"Can't open output data file %s\n",thestring);
  exit(1);
}

/* }}} */
      /* {{{ find max Z and X,Y,Z */

if (!coordset)
{
  X=Y=Z=0;
  maxz=0;

  for(int z=0; z<im.zsize(); z++)
    for(int y=0; y<im.ysize(); y++)
      for(int x=0; x<im.xsize(); x++)
	if ( (imz(x,y,z)>maxz) &&
	     ( (!haveclusters) ||   /* make max Z be inside a cluster if we found a cluster map */
	       (imclus(x,y,z)>0) && (!prewhiten || acs(x,y,z,1)!=0) ) )
	  {
	    maxz=imz(x,y,z);
	    X=x; Y=y; Z=z;
	  }
}
else
  maxz=imz(X,Y,Z);

/* }}} */
      /* {{{ create 1D data vector and find mean */

tsmean=0;

if(prewhiten)
     prewhiten_timeseries(acs.voxelts(X,Y,Z), im.voxelts(X,Y,Z), pwts, npts);
else
     pwts = im.voxelts(X,Y,Z);

for(t=0; t<npts; t++)
{
  data[t]=pwts(t+1);
  tsmean+=data[t];
}

tsmean/=npts;

/* }}} */
      /* {{{ output the data and model */

ymin=ymax=(int)data[0];

if (prewhiten)
     prewhiten_model(acs.voxelts(X,Y,Z), model, pwmodel, nevs, npts);

for(t=0; t<npts; t++)
{
  double fullmodel=0, copemodel=0;

  if (!modelfree)
    for(ev=0; ev<nevs; ev++)
      {
	if (prewhiten)
	  tmpf=pwmodel[t*nevs+ev]*impe[ev](X,Y,Z);
	else
	  tmpf=model[t*nevs+ev]*impe[ev](X,Y,Z);
	fullmodel += tmpf;
	if (j==0)
	  copemodel += tmpf*contrasts[i*nevs+ev];
      }
      
  fprintf(ofp,"%e ",data[t]);
  ymin=(int)MISCMATHS::Min(data[t],ymin);
  ymax=(int)MISCMATHS::Max(data[t],ymax);

  if (!modelfree)
    {
      if (j==0)
	{
	  fprintf(ofp,"%e ",tsmean+copemodel);
	  ymin=(int)MISCMATHS::Min(tsmean+copemodel,ymin);
	  ymax=(int)MISCMATHS::Max(tsmean+copemodel,ymax);
	}

      fprintf(ofp,"%e ",tsmean+fullmodel);
      ymin=(int)MISCMATHS::Min(tsmean+fullmodel,ymin);
      ymax=(int)MISCMATHS::Max(tsmean+fullmodel,ymax);

      if (j==0)
	fprintf(ofp,"%e ",data[t]-fullmodel+copemodel);
    }

  fprintf(ofp,"\n");
}

ymax+=(ymax-ymin)/5;
ymin-=(ymax-ymin)/20;

fclose(ofp);

/* }}} */
      /* {{{ call gnuplot */

sprintf(gpname,"%s/tsplot_%s%d",outputdir,statname,i+1);

if (!modelfree)
{
  if (j==0)
    /* {{{ full and partial model plots */

{

thestring2[0]=0;
for(ev=0;ev<nevs;ev++)
{
  if ((float)((int)contrasts[i*nevs+ev])==contrasts[i*nevs+ev])
    sprintf(thestring2,"%s %d",thestring2,(int)contrasts[i*nevs+ev]);
  else
    sprintf(thestring2,"%s %.1f",thestring2,contrasts[i*nevs+ev]);
}

sprintf(gpstart,"echo \"set term pbm color\nset size %f,0.5\nset title 'Contrast %d [%s ] :",MISCMATHS::Max(npts*4,600)/640.0,i+1,thestring2);

if (!coordset)
     sprintf(thestring3,"%s max Z stat of %.3f at voxel",gpstart,maxz);
else
     sprintf(thestring3,"%s Z stat of %.3f at selected voxel",gpstart,maxz);

sprintf(thestring,"%s (%d,%d,%d)'\nset yrange [%d:%d]\nplot '%s.txt' using 1 title 'data' with lines , '%s.txt' using 2 title 'cope partial model fit' with lines , '%s.txt' using 3 title 'full model fit' with lines \" | %sgnuplot | %sconvert - %s.gif",thestring3,X,Y,Z,ymin,ymax,gpname,gpname,gpname,binstr,binstr,gpname);

system(thestring);

sprintf(thestring,"%s (%d,%d,%d)'\nset yrange [%d:%d]\nplot '%s.txt' using 4 title 'reduced data' with lines , '%s.txt' using 2 title 'cope partial model fit' with lines \" | %sgnuplot | %sconvert - %sp.gif",thestring3,X,Y,Z,ymin,ymax,gpname,gpname,binstr,binstr,gpname);

system(thestring);

}

/* }}} */
  else
    /* {{{ full model plot */

{
  sprintf(gpstart,"echo \"set term pbm color\nset size %f,0.5\nset title 'F-test %d :",MISCMATHS::Max(npts*4,600)/640.0,i+1);

  if (!coordset)
    sprintf(thestring3,"%s max Z stat of %.3f at voxel",gpstart,maxz);
  else
    sprintf(thestring3,"%s Z stat of %.3f at selected voxel",gpstart,maxz);

  sprintf(thestring,"%s (%d,%d,%d)'\nset yrange [%d:%d]\nplot '%s.txt' using 1 title 'data' with lines , '%s.txt' using 2 title 'full model fit' with lines \" | %sgnuplot | %sconvert - %s.gif",thestring3,X,Y,Z,ymin,ymax,gpname,gpname,binstr,binstr,gpname);

  system(thestring);
}

/* }}} */
}
else
  /* {{{ data only plot */

{
  sprintf(gpstart,"echo \"set term pbm color\nset size %f,0.5\nset title '",MISCMATHS::Max(npts*4,600)/640.0);

  if (!coordset)
    sprintf(thestring3,"%s max Z stat of %.3f at voxel",gpstart,maxz);
  else
    sprintf(thestring3,"%s Z stat of %.3f at selected voxel",gpstart,maxz);

  sprintf(thestring,"%s (%d,%d,%d)'\nset yrange [%d:%d]\nplot '%s.txt' using 1 title 'data' with lines \" | %sgnuplot | %sconvert - %s.gif",thestring3,X,Y,Z,ymin,ymax,gpname,binstr,binstr,gpname);
 
  system(thestring);
}

/* }}} */

/* }}} */
      if (haveclusters)
	{
	  /* {{{ vars */

	  double *fullmodel=(double*)calloc(npts,sizeof(double)),
	    *copemodel=(double*)calloc(npts,sizeof(double)),
	    *data=(double*)calloc(npts,sizeof(double)),
	    wtotal=0;
	  int count=0;

/* }}} */
	  /* {{{ open data file */

sprintf(thestring,"%s/tsplotc_%s%d.txt",outputdir,statname,i+1);

if((ofp=fopen(thestring,"w"))==NULL)
{
  fprintf(stderr,"Can't open output data file %s\n",thestring);
  exit(1);
}

/* }}} */
	  /* {{{ create 1D data vector and find mean */

for(int x=0; x<im.xsize(); x++)
  for(int y=0; y<im.ysize(); y++)
    for(int z=0; z<im.zsize(); z++)
      if (imclus(x,y,z)>0 && (!prewhiten || acs(x,y,z,1)!=0))
      {
	count++;
	wtotal+=imw(x,y,z);

	if(prewhiten)
	  prewhiten_timeseries(acs.voxelts(x,y,z), im.voxelts(x,y,z), pwts, npts);
	else
	  pwts = im.voxelts(x,y,z);
	
	for(t=0; t<npts; t++)
	  data[t]+=pwts(t+1)*imw(x,y,z);
      }
 
/*printf("count=%d\n",count);*/

tsmean=0;
for(t=0; t<npts; t++)
{
  data[t]/=wtotal;
  tsmean+=data[t];
}
tsmean/=npts;

ymin=ymax=(int)data[0];

if (!modelfree)
  for(int x=0; x<im.xsize(); x++)
    for(int y=0; y<im.ysize(); y++)
      for(int z=0; z<im.zsize(); z++)
        if (imclus(x,y,z)>0 && (!prewhiten || acs(x,y,z,1) != 0))
        {
	  if (prewhiten)
	    prewhiten_model(acs.voxelts(x,y,z), model, pwmodel, nevs, npts);

	  for(t=0; t<npts; t++)
	    for(ev=0; ev<nevs; ev++)
	      {
		if (prewhiten)
		  tmpf=pwmodel[t*nevs+ev]*impe[ev](x,y,z);
		else
		  tmpf=model[t*nevs+ev]*impe[ev](x,y,z);
		fullmodel[t] += tmpf*imw(x,y,z);

		if (j==0)
		  copemodel[t] += tmpf*contrasts[i*nevs+ev]*imw(x,y,z);
	      }
	}

for(t=0; t<npts; t++)
{
  if (!modelfree)
    {
      fullmodel[t] /= wtotal;
      if (j==0)
	copemodel[t] /= wtotal;
    }
  
  fprintf(ofp,"%e ",data[t]);
  ymin=(int)MISCMATHS::Min(data[t],ymin);
  ymax=(int)MISCMATHS::Max(data[t],ymax);

  if (!modelfree)
    {
      if (j==0)
	{
	  fprintf(ofp,"%e ",tsmean+copemodel[t]);
	  ymin=(int)MISCMATHS::Min(tsmean+copemodel[t],ymin);
	  ymax=(int)MISCMATHS::Max(tsmean+copemodel[t],ymax);
	}

      fprintf(ofp,"%e ",tsmean+fullmodel[t]);
      ymin=(int)MISCMATHS::Min(tsmean+fullmodel[t],ymin);
      ymax=(int)MISCMATHS::Max(tsmean+fullmodel[t],ymax);
	
      if (j==0)
	fprintf(ofp,"%e ",data[t]-fullmodel[t]+copemodel[t]);
    }

  fprintf(ofp,"\n");
}

ymax+=(ymax-ymin)/5;
ymin-=(ymax-ymin)/20;

fclose(ofp);

/* }}} */
	  /* {{{ call gnuplot */

sprintf(gpname,"%s/tsplotc_%s%d",outputdir,statname,i+1);

if (!modelfree)
{
  if (j==0)
    /* {{{ call gnuplot via system call */
{
  sprintf(thestring,"%s averaged over %d voxels'\nset yrange [%d:%d]\nplot '%s.txt' using 1 title 'data' with lines , '%s.txt' using 2 title 'cope partial model fit' with lines , '%s.txt' using 3 title 'full model fit' with lines \" | %sgnuplot | %sconvert - %s.gif",gpstart,count,ymin,ymax,gpname,gpname,gpname,binstr,binstr,gpname);
  system(thestring);

  sprintf(thestring,"%s averaged over %d voxels'\nset yrange [%d:%d]\nplot '%s.txt' using 4 title 'reduced data' with lines , '%s.txt' using 2 title 'cope partial model fit' with lines \" | %sgnuplot | %sconvert - %sp.gif",gpstart,count,ymin,ymax,gpname,gpname,binstr,binstr,gpname);
  system(thestring);
}

/* }}} */
  else
    /* {{{ call gnuplot via system call */

{
  sprintf(thestring,"%s averaged over %d voxels'\nset yrange [%d:%d]\nplot '%s.txt' using 1 title 'data' with lines , '%s.txt' using 2 title 'full model fit' with lines \" | %sgnuplot | %sconvert - %s.gif",gpstart,count,ymin,ymax,gpname,gpname,binstr,binstr,gpname);
  system(thestring);
}

/* }}} */
}
else
  /* {{{ call gnuplot via system call */

{
  sprintf(thestring,"%s averaged over %d voxels'\nset yrange [%d:%d]\nplot '%s.txt' using 1 title 'data' with lines \" | %sgnuplot | %sconvert - %s.gif",gpstart,count,ymin,ymax,gpname,binstr,binstr,gpname);
  system(thestring);
}

/* }}} */

/* }}} */
	}
    }
}

/* }}} */
  /* {{{ clean up and exit */

  delete [] pwmodel;
  return 0;

/* }}} */
}

/* }}} */
