/*  mcflirt.cc - Motion Correction FLIRT
    
    Peter Bannister, FMRIB Image Analysis Group
    
    Copyright (C) 1999-2000 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */


#include <math.h>
#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <strstream>

#include "mjimage.h"
#include "generalio.h"
#include "interpolation.h"
#include "miscimfns.h"
#include "optimise.h"
#include "costfns.h"
#include "newmatap.h"
#include "vector.h"
#include "Globaloptions.h"
#include "Log.h"

using namespace MISCMATHS;
using namespace MISCIMFNS;
using namespace INTERPOLATION;
using namespace COSTFNS;
using namespace NEWMAT;
using namespace MJIMAGE;
using namespace GENERALIO;
using namespace UTILS;

//------------------------------------------------------------------------//

// OPTIMISATION SUPPORT


int vector2affine(const ColumnVector& params, int n, const ColumnVector& centre,
 		  Matrix& aff)
{
  if (n<=0) return 0;
  // order of parameters is 3 rotation + 3 translation + 3 scales + 3 skews
  // angles are in radians
  
  switch (Globaloptions::getInstance().anglerep) 
    {
    case Euler:
      compose_aff(params,n,centre,aff,construct_rotmat_euler);
      break;
    case Quaternion:
      compose_aff(params,n,centre,aff,construct_rotmat_quat);
      break;
    default:
      cerr << "Invalid Rotation Representation" << endl;
      return -1;
    }
  return 0;
}  


int vector2affine(const ColumnVector& params, int n, Matrix& aff)
{
  return vector2affine(params,n,Globaloptions::getInstance().impair->testvol.cog(),aff);
}


 int vector2affine(const float params[], int n, Matrix& aff)
 {
   ColumnVector vparams(12);
   for (int i=1; i<=n; i++) {
     vparams(i) = params[i];
   }
   return vector2affine(vparams,n,aff);
 }


 int affmat2vector(const Matrix& aff, int n, const ColumnVector& centre,
 		  ColumnVector& params)
 {
   switch (Globaloptions::getInstance().anglerep) 
     {
     case Euler:
       decompose_aff(params,aff,centre,rotmat2euler);
       break;
     case Quaternion:
       decompose_aff(params,aff,centre,rotmat2quat);
       break;
     default:
       cerr << "Invalid Rotation Representation" << endl;
     }
   return 0;
 }


 int affmat2vector(const Matrix& aff, int n, ColumnVector& params)
 {
   return affmat2vector(aff,n,Globaloptions::getInstance().impair->testvol.cog(),params);
 }


 void set_param_basis(Matrix &parambasis, int no_params)
 {
   parambasis = 0.0;
   for (int i=1; i<=no_params; i++) {
     parambasis(i,i)=1.0;
   }
 }

 void set_param_tols(ColumnVector &param_tol, int no_params)
 {
       // Tolerances are: 0.57 degrees (0.005 radians), 0.2mm translation
       //    0.02 scale and 0.001 skew
   float diagonal[12]={0.005, 0.005, 0.005, 0.2, 0.2, 0.2, 0.002, 0.002, 0.002,
   		      0.001, 0.001, 0.001};
   if (param_tol.Nrows()<no_params) {
     param_tol.ReSize(no_params);
   }
   for (int i=1; i<=no_params; i++) {
     param_tol(i)=diagonal[i-1];
   }
 }



 void initialise_params(ColumnVector& params)
 {
   Real paramsf[12] = {0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0};
   params << paramsf;
 }



 void powell_opt(ColumnVector& params, int no_params, ColumnVector& param_tol, 
 		int &no_its, float &fans, float (*costfunc)(const ColumnVector&), 
 		int itmax)
 {
   // sets up the initial parameters and calls the powell optimisation routine
   if (params.MaximumAbsoluteValue() < 0.001)  initialise_params(params);
   { 
     Matrix affmattst(4,4);
     vector2affine(params,no_params,affmattst);
   }
   Matrix parambasis(no_params,no_params);
   set_param_basis(parambasis,no_params);
   float ptol[13];
   for (int i=1; i<=no_params; i++) { ptol[i] = param_tol(i); }
  
   // the optimisation call
   //powell(params,parambasis,no_params,ptol,no_its, fans, costfunc, itmax);
   fans = MISCMATHS::optimise(params,no_params,param_tol,costfunc,no_its,itmax,
 			      Globaloptions::getInstance().boundguess);
 }


 void optimise(ColumnVector& params, int no_params, ColumnVector& param_tol, 
 	      int &no_its, float &fans, float (*costfunc)(const ColumnVector &), 
 	      int itmax=4)
 {
   powell_opt(params,no_params,param_tol,no_its,fans,costfunc,itmax);
 }

 ////////////////////////////////////////////////////////////////////////////

 // OPTIMISATION SUPPORT (cost function interfaces)


 float costfn(const Matrix& uninitaffmat)
 {
   Tracer tr("costfn");
   Matrix affmat = uninitaffmat * Globaloptions::getInstance().initmat;  // apply initial matrix
   float retval = 0.0;
   Globaloptions& globalopts = Globaloptions::getInstance();

   switch (globalopts.maincostfn) 
     {
     case NormCorr:
       retval = 1.0 - fabs(normcorr(globalopts.impair,affmat));  // MAXimise corr
       break;
     case CorrRatio:
       if (globalopts.smoothsize > 0.0) {
 	retval = 1.0 - corr_ratio_smoothed(globalopts.impair,affmat);  // MAXimise corr
       } else {
 	retval = 1.0 - corr_ratio(globalopts.impair,affmat);  // MAXimise corr
       }
       break;
     case Woods:
       retval = woods_fn(globalopts.impair,affmat);  // minimise variance/mean
       break;
     case MutualInfo:
       retval = -mutual_info(globalopts.impair,affmat);  // MAXimise info
       break;
     case NormMI:
       retval = -normalised_mutual_info(globalopts.impair,affmat);  // MAXimise
       break;
     default:
       cerr << "Invalid cost function type" << endl;
       break;
     }
   return retval;
 }


 float costfn(const ColumnVector& params)
 {
   Tracer tr("costfn");
   Matrix affmat(4,4);
   vector2affine(params,Globaloptions::getInstance().no_params,affmat);
   float retval = costfn(affmat);
   return retval;
 }
  
 /*
 float costfn(float params[])
 {
   Tracer tr("costfn");
   Matrix affmat(4,4);
   vector2affine(params,Globaloptions::getInstance().no_params,affmat);
   float retval = costfn(affmat);
   if (Globaloptions::getInstance().verbose>=5) {
     cout << Globaloptions::getInstance().impair->count++ << " : ";
     cout << retval << " :: ";
     for (int i=1; i<=Globaloptions::getInstance().no_params; i++) cout << params[i] << " ";
     cout << endl;
   }
   return retval;
 }
 */

 //------------------------------------------------------------------------//

 float estimate_scaling(const volume& vol) {
   Tracer tr("estimate_scaling");
   return Min(Min(vol.getx(),vol.gety()),vol.getz());
 }

 float estimate_scaling() {
   Tracer tr("estimate_scaling");
   return estimate_scaling(Globaloptions::getInstance().impair->refvol);
 }

 ////////////////////////////////////////////////////////////////////////////

 int optimise_strategy1(Matrix& matresult, float& fans, int input_dof, 
 		       int max_iterations)
 {
   Tracer tr("optimise_strategy1");
   // the most basic strategy - just do a single optimisation run at the
   //  specified dof
   int dof=input_dof;
   if (dof<6) { 
     cerr << "Erroneous dof " << dof << " : using 6 instead\n"; 
     dof=6; 
   }
   if (dof>12) {
     cerr << "Erroneous dof " << dof << " : using 12 instead\n"; 
     dof=12;
   }

   ColumnVector params(12), param_tol(12);
   int no_its=0;
   Globaloptions::getInstance().no_params = dof;
   set_param_tols(param_tol,12);
   param_tol = param_tol * estimate_scaling() * 0.2;
   affmat2vector(matresult,dof,params);
   optimise(params,dof,param_tol,no_its,fans,costfn,max_iterations);
   vector2affine(params,dof,matresult);
   return no_its;
 }  

 ////////////////////////////////////////////////////////////////////////////


 void usroptimise(Matrix &matresult, int usrdof, int usrmaxitn)
 {
   Tracer tr("usroptimise");
   // OPTIMISE
   int dof = Min(Globaloptions::getInstance().dof,usrdof);
  
   float costval=0.0;
   optimise_strategy1(matresult,costval,dof,usrmaxitn);
  
   // matresult is the desired solution (cost is costval)
 }


 void usrsetscale(volume& newrefvol, volume& newtestvol, int usrscale) {
   Tracer tr("usrsetscale");
   imagepair *globalpair=0;
   globalpair = new imagepair(newrefvol,newtestvol);
   globalpair->set_no_bins(Globaloptions::getInstance().no_bins/usrscale);
   Globaloptions::getInstance().impair = globalpair;
 }


 ////////////////////////////////////////////////////////////////////////////

 void correct(const int direction, volume& reference_volume, tvolume& timeseries, const float scaling) {
   Tracer tr("correct");
   Globaloptions& globalopts = Globaloptions::getInstance();
   volume testvol;

   Matrix offsettrans(4,4);
   Identity(offsettrans);
   vector<Matrix> mat_array(globalopts. no_volumes,offsettrans);

   Log& logger = Log::getInstance();
  
   int i=globalopts. refnum + direction;

   while ( i != (direction == 1 ? globalopts. no_volumes : -1)) {
     cerr << "Registering volume [" << i << "]" << endl;
     testvol = timeseries[i];
  
     if (globalopts. edgeflag){
       cerr << "Calculating contour image for volume [" << i << "]" << endl;
       fixed_edge_detect(testvol, 15000);  
     } else if (globalopts. gdtflag){
       cerr << "Calculating gradient image for volume [" << i << "]" << endl;
       volume gtempvol = testvol;
       gradient(gtempvol, testvol);
     }
    
     usrsetscale(reference_volume,testvol,(int) scaling);
     usroptimise(mat_array[i],globalopts.dof,1);
     offsettrans = mat_array[i];

     i += direction;
     if (i < globalopts. no_volumes || i >= 0 )
       mat_array[i] = offsettrans; // reinitialise next volume with previous estimate
   }
 
   if (scaling == 2){ // final sampling so apply transformation matricies to timeseries
     for (int i=0; i< globalopts. no_volumes; i++) {
       testvol = timeseries[i];
       filled_affine_transform(testvol, timeseries[i], mat_array[i]*globalopts. initmat);
     }
   }
 }

 void run_and_save_stats(const tvolume& timeseries) {
   Tracer tr("run_and_save_stats");
   Globaloptions& globalopts = Globaloptions::getInstance();
   volume variancevol, meanvol, sigmavol;

   int vmax = timeseries.getSize();
   meanvol = timeseries[0];
   variancevol = timeseries[0];
   sigmavol = timeseries[0];

   meanvol.reinitialize(timeseries[0].xsize(), timeseries[0].ysize(), (timeseries[0].zsize() - 2));
   variancevol.reinitialize(meanvol);
   sigmavol.reinitialize(meanvol);

   meanvol = 0.0;
   variancevol = 0.0;
   sigmavol = 0.0;

   // calculate the mean value and variance at each voxel
   for (int x=0; x< timeseries[0].xsize(); x++) {
     for (int y=0; y< timeseries[0].ysize(); y++) {
       for (int z=0; z< (timeseries[0].zsize()-2); z++) {
 	for (int i=0; i< vmax; i++) 
 	  meanvol(x,y,z) = meanvol(x,y,z) + timeseries[i](x,y,z+1);
 	meanvol(x,y,z) = meanvol(x,y,z)/(float)vmax;
 	for (int i=0; i< vmax; i++)
 	  variancevol(x,y,z) = variancevol(x,y,z) + (timeseries[i](x,y,z+1) - meanvol(x,y,z))*(timeseries[i](x,y,z+1) - meanvol(x,y,z));
 	variancevol(x,y,z) = variancevol(x,y,z)/((float)(vmax - 1));
 	sigmavol(x,y,z) = sqrt(variancevol(x,y,z));
       }
     }
   }

   cerr << "Saving mean volume... " << endl;
   save_volume(meanvol, "mean_"+globalopts.outputfname);
  
   cerr << "Saving variance volume... " << endl;
   save_volume(variancevol, "variance_"+globalopts.outputfname);

   cerr << "Saving standard deviation volume... " << endl;
   save_volume(sigmavol, "sigma_"+globalopts.inputfname);
 }



int main (int argc,char** argv)
{
  Tracer tr("main");
  tvolume timeseries;
  volume refvol, anisorefvol;
  Globaloptions& globalopts = Globaloptions::getInstance();
  short dtype;
  Log& logger = Log::getInstance();
  
  cerr << 
    endl << "McFLIRT v 0.1.3 - Linear motion correction tool" << 
    endl << "This software is currently alpha." << 
    endl << "Please send any comments/ queries to: prb@fmrib.ox.ac.uk" << 
    endl << endl;
  
  globalopts.parse_command_line(argc, argv);
 
  logger.establishDir("mats");

  cerr << "Reading time series... " << endl;
  read_series(timeseries, globalopts.inputfname, dtype);
  globalopts.datatype = dtype;
  
  globalopts. no_volumes = timeseries.getSize();
  if (globalopts. refnum == -1)
    globalopts. refnum = globalopts. no_volumes/2;

  for (float current_scale=8.0; current_scale > 1.9; current_scale /= 2.0) {
    anisorefvol = timeseries[globalopts. refnum];
    cerr << "Rescaling reference volume [" << globalopts. refnum << "] to " 
	 << current_scale << " mm pixels" << endl;
    isotropic_rescale(refvol,anisorefvol,current_scale);
    
    if (globalopts. edgeflag){
      cerr << "Calculating contour image for reference volume" << endl;
      fixed_edge_detect(refvol, 15000);  
      cerr << "Saving contour reference volume... " << endl;
      save_volume(refvol, "crefvol_"+globalopts.outputfname);
    } else if (globalopts. gdtflag){
      cerr << "Calculating gradient image for reference volume" << endl;
      volume gtempvol = refvol;
      gradient(gtempvol, refvol);
      
      cerr << "Saving gradient reference volume... " << endl;
      save_volume(refvol, "grefvol_"+globalopts.outputfname);
    }
  
    Identity(globalopts.initmat);

    correct(1, refvol, timeseries, current_scale);
    correct(-1, refvol, timeseries, current_scale);
  }
  
  cerr << "Saving motion corrected time series... " << endl;
  save_series(timeseries, globalopts.outputfname, globalopts.datatype);

  run_and_save_stats(timeseries);
}








