/*  MELODIC - Multivariate exploratory linear optimized decomposition into 
              independent components
    
    melodicdata.cc - data handler / container class

    Christian F. Beckmann, FMRIB Image Analysis Group
    
    Copyright (C) 1999-2001 University of Oxford */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

#include "newimageall.h"
#include "Log.h"
#include "melodicoptions.h"
#include "melodicdata.h"
#include "melodic.h"
#include "Log.h"
#include <time.h>

using namespace Utilities;
using namespace NEWIMAGE;

namespace Melodic{

  void MelodicData::setup()
  {
    
    {volume4D<float> RawData;
    read_volume4D(RawData,opts.inputfname.value());
    for(int ctr=1; ctr<=opts.dummy.value(); ctr++){
      RawData.deletevolume(ctr);
    }    
    
    // calculate a Mean image and save it
    Mean = meanvol(RawData);
    tmpnam(Mean_fname); // generate a tmp name
    save_volume(Mean,Mean_fname);

    create_mask(RawData, Mask);
    Data = RawData.matrix(Mask);

    //mask out constant voxels
    message(" Exluding voxels with constant value " << endl);
    Matrix DStDev=stdev(Data);
    volume4D<float> tmpMask;
    tmpMask.setmatrix(DStDev,Mask);
    float tMmax;
    volume<float> tmpMask2;
    tmpMask2 = tmpMask[0];
    tMmax = tmpMask2.max();
    Mask = threshold(tmpMask2,(float) 0.0001,tMmax);
    Data = RawData.matrix(Mask);}

    message(" Data size : " << Data.Nrows() << " x " << Data.Ncols() <<endl);

    //segment the mean image and get GM probabilities
    if(opts.segment.value()){
      volume4D<float> GMtmp;
      create_GMseg(GMtmp);
      GMseg = GMtmp.matrix(Mask).Row(1);
      message("done" << endl);
    }
    
    // clean /tmp
    char callRMstr[1000];
    ostrstream osc(callRMstr,1000);
    osc  << "rm " << string(Mean_fname) <<"*  " << '\0';
    system(callRMstr);

    // remove mean volume
    if(opts.remove_meanvol.value()||opts.varnorm.value()){
      message(string("Remove mean image ... "));
      meanR=mean(Data);
      Data=remmean(Data); 
      message("done" << endl);
    }

    // variance-normalize the data
    if(opts.varnorm.value()){
      message(string("Perform variance-normalization ... "));
      Matrix Data_stdev = pow(stdev(Data),-1);
      Data = SP(Data,ones(Data.Nrows(),1)*Data_stdev);   
      message("done"<<endl);
    }
 
    // remove row mean
    // message(string("Remove mean time course ... "));
    meanC=mean(Data,2);
    Data=remmean(Data.t()).t(); 
    // message("done"<<endl);

    //seed the random number generator
    double tmptime = time(NULL);
    srand((unsigned int) tmptime/Data.Ncols()*Data.Nrows());

  } // void setup()

  void MelodicData::save()
  {

    message("Writing results to : " << endl);
    //Output IC	
    if(IC.Storage()>0){
      volume4D<float> tempVol;	
      volumeinfo tempInfo;
      read_volume4D(tempVol,opts.inputfname.value(),tempInfo); 
      tempVol.setmatrix(IC,Mask);
      strncpy(tempInfo.header.hist.aux_file,"render3",24);
      save_volume4D(tempVol,logger.appendDir(opts.outputfname.value() 
					     + "_IC"),tempInfo);
      message("  " << logger.appendDir(opts.outputfname.value() + "_IC") <<endl);
    }

    //Output mixMatrix
    if(mixMatrix.Storage()>0){
      write_ascii_matrix(logger.appendDir(opts.outputfname.value() + "_mix"),
			 mixMatrix); 
      mixFFT=calc_FFT(mixMatrix);
      write_ascii_matrix(logger.appendDir(opts.outputfname.value() + "_FTmix"),
			 mixFFT);      
      message("  "<<
	      logger.appendDir(opts.outputfname.value() + "_mix") <<endl);
      message("  "<<
	      logger.appendDir(opts.outputfname.value() + "_FTmix") <<endl);
    }

    //Output unmixMatrix
    if(opts.output_unmix.value() && unmixMatrix.Storage()>0){
      write_ascii_matrix(opts.outputfname.value() + "_unmix",unmixMatrix);
      message("  "<<
	  logger.appendDir(opts.outputfname.value() + "_unmix") <<endl);
    }

    //Output Mask
    if(opts.output_mask.value()){
      save_volume(Mask,logger.appendDir("melodic_mask"));
      message("  "<<
	  logger.appendDir("melodic_mask") <<endl);
    }

    //Output mean
    if(opts.output_mean.value() && meanC.Storage()>0 && meanR.Storage()>0){
      write_ascii_matrix(logger.appendDir(opts.outputfname.value() + "_meanR"),
			 meanR);
      write_ascii_matrix(logger.appendDir(opts.outputfname.value() + "_meanC"),
			 meanC);
      message("  "<<
	  logger.appendDir(opts.outputfname.value() + "_meanR") <<endl);
      message("  "<<
	  logger.appendDir(opts.outputfname.value() + "_meanC") <<endl);
    }

    //Output white
    if(opts.output_white.value() && whiteMatrix.Storage()>0&&
       dewhiteMatrix.Storage()>0){
      write_ascii_matrix(logger.appendDir(opts.outputfname.value() + "_white"),
			 whiteMatrix);
      write_ascii_matrix(logger.appendDir(opts.outputfname.value() + "_dewhite"),
			 dewhiteMatrix);
      message("  "<<
	  logger.appendDir(opts.outputfname.value() + "_white") <<endl);
      message("  "<<
	  logger.appendDir(opts.outputfname.value() + "_dewhite") <<endl);
    }

    //Output PCA
    if(opts.output_pca.value() && pcaD.Storage()>0&&pcaE.Storage()>0){
      write_ascii_matrix(logger.appendDir(opts.outputfname.value() + "_pcaE"),
			 pcaE);
      write_ascii_matrix(logger.appendDir(opts.outputfname.value() + "_pcaD"),
			 (Matrix) diag(pcaD));
      message("  "<<
	  logger.appendDir(opts.outputfname.value() + "_pcaE") <<endl);
      message("  "<<
	  logger.appendDir(opts.outputfname.value() + "_pcaD") <<endl);
    }
  } //void save()
  
  int MelodicData::remove_components()
  {  
    string tmpstr;
    cout << endl << "  Components to delete: ";
    getline(cin,tmpstr);

    Matrix noiseMix;
    Matrix noiseIC;
   
    int ctr=0;    
    char *p;
    char t[1024];
    const char *discard = ", [];{(})abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ~!@#$%^&*_-=+|\':><./?";

    strcpy(t, tmpstr.c_str());
    p=strtok(t,discard);
    ctr = atoi(p);
    noiseMix = mixMatrix.Column(ctr);
    noiseIC  = IC.Row(ctr).t();    

    do{
      p=strtok(NULL,discard);
      if(p){
	ctr = atoi(p);
        if(ctr>0 && ctr<=mixMatrix.Ncols()){
	  noiseMix |= mixMatrix.Column(ctr);
	  noiseIC  |= IC.Row(ctr).t();
	}
      }
    }while(p);

    cout << endl << "  Components removed" << endl;
    
    Matrix newData;
    newData = Data - noiseMix * noiseIC.t() + 
      ones(Data.Nrows(),1)*meanR + meanC*ones(1,Data.Ncols());
    
    volume4D<float> tmp;
    read_volume4D(tmp,opts.inputfname.value()); 
    tmp.setmatrix(newData,Mask);
    save_volume4D(tmp,logger.appendDir(opts.outputfname.value() + "_ICAfiltered")); 
    cout << "  Written new data set to "<< 
      logger.appendDir(opts.outputfname.value() + "_ICAfiltered") << endl;
    return 0;
  } // int remove_components()

  void MelodicData::create_GMseg(volume4D<float> &GMtmp)
  {
    message("Create segmentation mask ... ");
    string FAST_fname = string(Mean_fname)+"_brain";
    cout << FAST_fname << endl;
    string FAST_path = "$FSLDIR/bin/fast";
    string FAST_optarg = "-n -op -v0 "; // see FAST
    string GMseg_fname = FAST_fname+"_pbmap_1.hdr";
    
    // Setup external call to FAST:
    
    char callFASTstr[1000];
    ostrstream osc(callFASTstr,1000);
    osc  << FAST_path << " " << FAST_optarg << string(FAST_fname)+".hdr" << 
      " 2> /dev/null " << '\0';
    
    message(" Calling FAST: " << callFASTstr << endl);
    system(callFASTstr);

    message("  reading back : " << GMseg_fname<< endl);
 
    read_volume4D(GMtmp,GMseg_fname);
  } 
 
  void MelodicData::create_mask(volume4D<float> &theData, 
				volume<float> &theMask)
  {
    if(opts.use_mask.value() && opts.maskfname.value().size()>0){   // mask provided 
      read_volume(theMask,opts.maskfname.value());
      message("Mask provided : " << opts.maskfname.value()<<endl);
    }
    else{
      if(opts.perf_bet.value() && opts.use_mask.value()){ //use BET
	message("Create mask ... ");
	// set up all strings
	string BET_outputfname = string(Mean_fname)+"_brain";
	string BET_path = "$FSLDIR/bin/bet";
	string BET_optarg = "-m -f 0.4"; // see man bet
	string Mask_fname = BET_outputfname+"_mask.hdr";

	// Setup external call to BET:

	char callBETstr[1000];
	ostrstream osc(callBETstr,1000);
	osc  << BET_path << " " << Mean_fname << " " 
	     << BET_outputfname << " " << BET_optarg << " > /dev/null " << '\0';
	
        message("  Calling BET: " << callBETstr << endl);
	system(callBETstr);
	
	// read back the Mask file   
	read_volume(theMask,Mask_fname);

	message("done" << endl);
      }  
      else{
	if(opts.use_mask.value()){   //just threshold the Mean
	  message("Create mask ... ");
	  float Mmin, Mmax;
	  Mmin = Mean.min(); Mmax = Mean.max();
	  theMask = threshold(Mean,Mmin + opts.threshold.value()* (Mmax-Mmin),Mmax);
	  message("done" << endl);
	}
	else{ //well, don't threshold then
	  theMask = Mean;
	  theMask = 1.0;
	}
      }
    }
    if(opts.remove_endslices.value()){ 
      // just in case mc introduced something nasty
      message("  Deleting end slices" << endl);
      for(int ctr1=theMask.miny(); ctr1<=theMask.maxy(); ctr1++){
	for(int ctr2=theMask.minx(); ctr2<=theMask.maxx(); ctr2++){   
	  theMask(ctr2,ctr1,Mask.minz()) = 0.0;
	  theMask(ctr2,ctr1,Mask.maxz()) = 0.0;
	}
      }
    }
  } //void create_mask()

  void MelodicData::remove_comp(Matrix& Mat, const int numIC)
  {
    for(int ctr=1; ctr<=Mat.Nrows(); ctr++){
      Mat(ctr,numIC)=0.0;
    }
  } //void remove_comp


  void MelodicData::status(const string &txt)
  {
    cout << "MelodicData Object " << txt << endl;
    if(Data.Storage()>0){cout << "Data: " << Data.Nrows() <<"x" << Data.Ncols() << endl;}else{cout << "Data empty " <<endl;}
    if(pcaE.Storage()>0){cout << "pcaE: " << pcaE.Nrows() <<"x" << pcaE.Ncols() << endl;}else{cout << "pcaE empty " <<endl;}
    if(pcaD.Storage()>0){cout << "pcaD: " << pcaD.Nrows() <<"x" << pcaD.Ncols() << endl;}else{cout << "pcaD empty " <<endl;}
    if(whiteMatrix.Storage()>0){cout << "white: " << whiteMatrix.Nrows() <<"x" << whiteMatrix.Ncols() << endl;}else{cout << "white empty " <<endl;}
    if(dewhiteMatrix.Storage()>0){cout << "dewhite: " << dewhiteMatrix.Nrows() <<"x" << dewhiteMatrix.Ncols() << endl;}else{cout << "dewhite empty " <<endl;}
    if(mixMatrix.Storage()>0){cout << "mix: " << mixMatrix.Nrows() <<"x" << mixMatrix.Ncols() << endl;}else{cout << "mix empty " <<endl;}
    if(unmixMatrix.Storage()>0){cout << "unmix: " << unmixMatrix.Nrows() <<"x" << unmixMatrix.Ncols() << endl;}else{cout << "unmix empty " <<endl;}
    if(IC.Storage()>0){cout << "IC: " << IC.Nrows() <<"x" << IC.Ncols() << endl;}else{cout << "IC empty " <<endl;}
   
  } //void status()

  Matrix MelodicData::calc_FFT(Matrix& Mat)
  {
    Matrix res;
    for(int ctr=1; ctr <= Mat.Ncols(); ctr++)
      {
	ColumnVector tmpCol;
	tmpCol=Mat.Column(ctr);
	ColumnVector FtmpCol_real;
	ColumnVector FtmpCol_imag;
	ColumnVector tmpPow;
	if(tmpCol.Nrows()%2 != 0){
	  Matrix empty(1,1); empty=0;
	  tmpCol &= empty;}
	RealFFT(tmpCol,FtmpCol_real,FtmpCol_imag);
	tmpPow = pow(FtmpCol_real,2)+pow(FtmpCol_imag,2);
	tmpPow = tmpPow.Rows(2,tmpPow.Nrows());
	if(opts.logPower.value()) tmpPow = log(tmpPow);
	if(res.Storage()==0){res= tmpPow;}else{res|=tmpPow;}
      }
    return res;
  } //Matrix calc_FFT()

  Matrix MelodicData::smoothColumns(const Matrix &inp)
  {
    Matrix temp(inp);
    int ctr1 = temp.Nrows();
    Matrix temp2(temp);
    temp2=0;
    
    temp = temp.Row(4) & temp.Row(3) & temp.Row(2) & temp & temp.Row(ctr1-1) 
      & temp.Row(ctr1-2) &temp.Row(ctr1-3);
    
    double kern[] ={0.0045 , 0.055, 0.25, 0.4, 0.25, 0.055, 0.0045};
    double fac = 0.9090909;

    for(int cc=1;cc<=temp2.Ncols();cc++){
      for(int cr=1;cr<=temp2.Nrows();cr++){
	temp2(cr,cc) = fac*( kern[0] * temp(cr,cc) + kern[1] * temp(cr+1,cc) + 
			     kern[2] * temp(cr+2,cc) + kern[3] * temp(cr+3,cc) + 
			     kern[4] * temp(cr+4,cc) + kern[5] * temp(cr+5,cc) + 
			     kern[6] * temp(cr+6,cc));
      }
    }
    return temp2;
  }
}
