
/*  MELODIC - Multivariate exploratory linear optimized decomposition into 
              independent components
    
    melodicoptions.h - class for command line options

    Christian F. Beckmann, FMRIB Image Analysis Group
    
    Copyright (C) 1999-2001 University of Oxford */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */


#ifndef __MELODICOPTIONS_h
#define __MELODICOPTIONS_h

#include <string>
#include <strstream>
#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <stdio.h>
#include "options.h"
#include "Log.h"
#include "melodic.h"


using namespace Utilities;

namespace Melodic {

class MelodicOptions {
 public:
  static MelodicOptions& getInstance();
  ~MelodicOptions() { delete gopt; }
  
  string version;
  
  Option<string> logdir;
  Option<string> inputfname;
  Option<string> outputfname;
  Option<string> guessfname;
  Option<string> maskfname;
  Option<string> paradigmfname;
  Option<string> nonlinearity;
  Option<string> approach;
  
  Option<int>  pca_dim;
  Option<bool> segment;
  Option<int>  numICs;
  Option<int>  maxNumItt;
  Option<int>  maxRestart;
  Option<int>  dummy;
  Option<bool> varnorm;
  Option<int>  repeats;
  Option<float> nlconst1;
  Option<float> nlconst2;
  Option<float> epsilon;
  Option<float> threshold;


  Option<bool> help;
  Option<bool> verbose;  
  Option<bool> vers;
  Option<bool> copyright;
  Option<bool> perf_bet;

  Option<bool> remove_meanvol;
  Option<bool> remove_endslices;
  Option<bool> output_pca;
  Option<bool> output_white;
  Option<bool> output_mask;
  Option<bool> output_mean;
  Option<bool> use_mask;
  Option<bool> output_unmix;
  Option<bool> guess_remderiv;
  Option<bool> filter; 
  Option<bool> logPower;
  Option<bool> tsmooth;

  int retrystep;

  void parse_command_line(int argc, char** argv, Log& logger,  const string &p_version);

 private:
  MelodicOptions();  
  const MelodicOptions& operator=(MelodicOptions&);
  MelodicOptions(MelodicOptions&);

  OptionParser options; 
      
  static MelodicOptions* gopt;

  void print_usage(int argc, char *argv[]);  
  void print_version(); 
  void print_copyright();
  void status();
};

 inline MelodicOptions& MelodicOptions::getInstance(){
   if(gopt == NULL)
     gopt = new MelodicOptions();
   
   return *gopt;
 }

 inline MelodicOptions::MelodicOptions() :
   logdir(string("-o,--outdir"), string("melodic.log"),
	  string("output directory name"), 
	  false, requires_argument),
   inputfname(string("-i,--in"), string(""),
	      string("input file name"), 
	      true, requires_argument),
   outputfname(string("-O,--out"), string("melodic"),
	   string("output file name"), 
	   false, requires_argument,false),
   guessfname(string("-g,--guess"), string(""),
	   string("file name of guess of mixing matrix"), 
	   false, requires_argument, false),
   maskfname(string("-m,--mask"), string(""),
	   string("file name of mask for thresholding"), 
	   false, requires_argument),
   paradigmfname(string("-p,--paradigm"),  string(""),
	   string("file name of FEAT paradigm file"), 
	   false, requires_argument),
   nonlinearity(string("--nl,--nonlinearity"), string("tanh"),
	   string("nonlinearity: gauss, tanh, pow3"), 
	   false, requires_argument),
   approach(string("-a,--approach"),  string("symm"),
	   string("approach for ICA estimation: defl, symm"), 
	   false, requires_argument),
  
   pca_dim(string("-d,--dim"), -1,
	   string("dimensionality reduction into #num dimesnions (PCA)"), 
	   false, requires_argument),
   segment(string("-s,--segment"), false,
	   string("use gray-matter probabilities for spatial covariance matrix"),
	   false, no_argument,false),
   numICs(string("-n,--numICs"), -1,
	   string("numer of IC's to extract"), 
	   false, requires_argument),
   maxNumItt(string("--maxItt"),  500,
	   string("maximum number of itterations before restart"), 
	   false, requires_argument),
   maxRestart(string("--maxRestarts"),  6,
	   string("maximum number of restarts"), 
	   false, requires_argument),
   dummy(string("--dummy"),  0,
	   string("number of dummy volumes"), 
	   false, requires_argument),
   varnorm(string("--vn,--varnorm"), true,
	   string("switch off variance normalisation"), 
	   false, no_argument),
   repeats(string("--repeats"), 1,
	   string("number of repeats (multistart)"), 
	   false, requires_argument),
   nlconst1(string("--nl1,--nlconst1"),  1.0,
	   string("nonlinear constant 1"), 
	   false, requires_argument),
   nlconst2(string("--nl2,--nlconst2"),  1.0,
	   string("nonlinear constant 2"), 
	   false, requires_argument),
   epsilon(string("--eps,--epsilon"), 0.0005,
	   string("minimum error change"), 
	   false, requires_argument),
   threshold(string("-t,--threshold"),  0.10,
	   string("brain/ non-brain threshold (if --nobet selected)"), 
	   false, requires_argument),

   help(string("-h,--help"),  false,
	   string("prints this help message"), 
	   false, no_argument),
   verbose(string("-v,--verbose"), false,
	   string("switch on diagnostic messages"), 
	   false, no_argument),
   vers(string("-V,--version"), false,
	   string("prints version information"), 
	   false, no_argument),
   copyright(string("--copyright"), false,
	   string("prints copyright information"), 
	   false, no_argument),
   perf_bet(string("--nobet"), true,
	   string("switch off BET"), 
	   false, no_argument),
   remove_meanvol(string("--keepmeanvol"), true,
	   string("do not subtract mean volume"), 
	   false, no_argument),
   remove_endslices(string("--remEndslices"),  false,
	   string("delete end slices (motion correction artefacts)"), 
	   false, no_argument),
   output_pca(string("--Opca"),  false,
	   string("output PCA matrices"), 
	   false, no_argument),
   output_white(string("--Owhite"),  false,
	   string("output whitening/dewhitening matrices"), 
	   false, no_argument),
   output_mask(string("--Omask"),  true,
	   string("do not output mask volume"), 
	   false, no_argument),
   output_mean(string("--Omean"),  false,
	   string("output mean volume"), 
	   false, no_argument),
   use_mask(string("--nomask"), true,
	   string("switch off masking"), 
	   false, no_argument),
   output_unmix(string("--Ounmix"),  false,
	   string("output unmixing matrix"), 
	   false, no_argument),
   guess_remderiv(string("--remderiv"),  false,
	   string("removes every second entry in paradigm file (EV derivatives)"), 
	   false, no_argument, false),
   filter(string("-f,--filter"),  false,
	   string("interactive removal of components"), 
	   false, no_argument, false),
   logPower(string("--logPower"),  false,
	   string("calculate log of power for frequency spectrum"), 
	    false, no_argument, false),
   tsmooth(string("--spca"),  false,
	   string("smooth the eigenvectors prior to ICA"), 
	    false, no_argument),
   retrystep(3),
   options(string(""), " melodic <options> -i <filename>\n melodic --help")
   {
     try {  
            options.add(logdir);
	    options.add(inputfname);
	    options.add(outputfname);
	    options.add(guessfname);
	    options.add(maskfname);
	    options.add(paradigmfname);
	    options.add(nonlinearity);
	    options.add(approach);
  
	    options.add(pca_dim);
	    options.add(segment);
	    options.add(numICs);
	    options.add(maxNumItt);
	    options.add(maxRestart);
	    options.add(dummy);
	    options.add(varnorm);
	    options.add(repeats);
	    options.add(nlconst1);
	    options.add(nlconst2);
	    options.add(epsilon);
	    options.add(threshold);
	    
	    options.add(help);
	    options.add(verbose);
	    options.add(vers);
	    options.add(copyright);
	    options.add(perf_bet);
	    options.add(remove_meanvol);
	    options.add(remove_endslices);
	    options.add(output_pca);
	    options.add(output_white);
	    options.add(output_mask);
	    options.add(output_mean);
	    options.add(use_mask);
	    options.add(output_unmix);
	    options.add(guess_remderiv);
	    options.add(filter);
	    options.add(logPower);
	    options.add(tsmooth);
     }
     catch(X_OptionError& e) {
       options.usage();
       cerr << endl << e.what() << endl;
     } 
     catch(std::exception &e) {
       cerr << e.what() << endl;
     }    
     
   }
}

#endif



