#include <ctime>
#include "fileps.h"
using namespace std;

/////////////////////////////////////////////////////////////////////
/// ZPSImage

static	char *HEX = "0123456789abcdef";

static	char RGBcolorimage[] = "\
/bwproc {\n\
    rgbproc\n\
    dup length 3 idiv string 0 3 0\n\
    5 -1 roll {\n\
	add 2 1 roll 1 sub dup 0 eq {\n\
	    pop 3 idiv\n\
	    3 -1 roll\n\
	    dup 4 -1 roll\n\
	    dup 3 1 roll\n\
	    5 -1 roll put\n\
	    1 add 3 0\n\
	} { 2 1 roll } ifelse\n\
    } forall\n\
    pop pop pop\n\
} def\n\
/colorimage where {pop} {\n\
    /colorimage {pop pop /rgbproc exch def {bwproc} image} bind def\n\
} ifelse\n\
";

/*
 * Adobe Photoshop requires a comment line of the form:
 *
 * %ImageData: <cols> <rows> <depth>  <main channels> <pad channels>
 *	<block size> <1 for binary|2 for HEX> "data start"
 *
 * It is claimed to be part of some future revision of the EPS spec.
 */
static void PhotoshopBanner(fstream& fs, UINT w, UINT h, int bs, int nc)
{
	fs << "%%ImageData: " << (long) w << ' ' << (long) h << " 8 " << nc << " 0 " << bs << " 2 \"";
	if(nc>1) fs << "false 3 colorimage"; else fs << "image";
	fs << "\"\n";
}

static void PSHead(fstream& fs, const char* filename, float pw, float ph, float ox, float oy)
{
    time_t t;

	t = time(0);
	fs << "%%!PS-Adobe-3.0 EPSF-3.0\n";
	fs << "%%Creator: Yongyue Zhang's ImageReader\n";
	fs << "%%Title: " << filename << "\n";
	fs << "%%CreationDate: " << ctime(&t);
	fs << "%%DocumentData: Clean7Bit\n";
	fs << "%%Origin: " << ox << ' ' << oy << "\n";
	/* NB: should use PageBoundingBox */
	fs << "%%BoundingBox: 0 0 " << (long) ceil(pw) << ' ' << (long) ceil(ph) << "\n",
	fs << "%%LanguageLevel: 1\n";
	fs << "%%Pages: (atend)\n%%EndComments\n%%BeginSetup\n%%EndSetup\n";
}

static void PSColorContigPreamble(const ZImageBase& image, fstream& fs, int nc)
{
	PhotoshopBanner(fs, image.Width(), image.Height(), 1, nc);
	fs << "/line " << image.BytesPerRow() << " string def\n";
	fs << image.Width() << ' ' << image.Height() << " 8\n";
	fs << "[" << image.Width() << " 0 0 -" << image.Height() << " 0 " << image.Height() << "]\n";
	fs << "{currentfile line readhexstring pop} bind\nfalse " << nc << " colorimage\n";
}

#define MAXLINE		36
#define	DOBREAK(len, howmany, fs) \
	if (((len) -= (howmany)) <= 0) {	\
		fs << '\n';			\
		(len) = MAXLINE-(howmany);	\
	}
#define	PUTHEX(c,fs)	fs << HEX[((c)>>4)&0xf] << HEX[(c)&0xf];

static void PSDataColorContig(const ZRGBByteImage& image, fstream& fs, int nc)
{
	int breaklen = MAXLINE;
	const ZRGB<BYTE>* p1 = image.ImageOrigin();
	for (UINT j=0; j<image.Height(); j++, p1+=image.ImageWidth()) 
	{
		const ZRGB<BYTE>* p2 = p1;
		for (UINT i=0; i<image.Width(); i++, p2++) 
		{
			DOBREAK(breaklen, 3, fs);
			PUTHEX(p2->r,fs); PUTHEX(p2->g,fs); PUTHEX(p2->b,fs);
		}
	}
}

static void PSDataBW(const ZGrayByteImage& image, fstream& fs)
{
	int breaklen = MAXLINE;
	
	for(ZGrayByteImage::const_iterator p(image); p.more(); p++)
	{
		DOBREAK(breaklen, 1, fs); PUTHEX(*p, fs);
	}
}

static void PSpage(const ZImageBase& image, fstream& fs)
{
	if(image.isColor())
	{
		fs << RGBcolorimage;
		PSColorContigPreamble(image, fs, 3);
		PSDataColorContig((const ZRGBByteImage&)image, fs, 3);
	}
	else
	{
		PhotoshopBanner(fs, image.Width(), image.Height(), 1, 1);
		fs << "/scanLine " << image.BytesPerRow() << " string def\n";
		fs << image.Width() << ' ' << image.Height() << " 8\n";
		fs << "[" << image.Width() << " 0 0 -" << image.Height() << " 0 " << image.Height() << "]\n";
		fs << "{currentfile scanLine readhexstring pop} bind\nimage\n";
		PSDataBW((const ZGrayByteImage&)image, fs);
	}
	fs << '\n';
}

ZImageBase*	ZPSImage::ReadFile()
{
	if(!IsOpen())
	{
		ZWarning("PSReader", "%s -- File is not opened!", m_filename.c_str());
		return 0;
	}

	bool color = false;
	ZImageBase* pImage = 0;
	char buf[1024]; buf[0] = 0;
	while(!m_ImageStream.eof())
	{
		m_ImageStream.getline(buf, 1024);
		if(!strncmp(buf, "/scanLine", 9)) break;
		if(!strncmp(buf, "/line", 5)) { color = true; break; }
	}
	if(m_ImageStream.eof()) 
	{
		ZWarning("PSReader", "%s -- Not a valid PS file!", m_filename.c_str());
		return 0;
	}

	int width, height, bit;
	m_ImageStream >> width >> height >> bit;
	if(bit != 8)
	{
		ZWarning("PSReader", "%s -- Not a valid PS file!", m_filename.c_str());
		return 0;
	}

	for(int l=0; l<4; l++)	m_ImageStream.getline(buf, 1024);

	if(color) pImage = new ZRGBByteImage(width, height); 
	else pImage = new ZGrayByteImage(width, height); 

	int nb = color ? 3 : 1; int size = pImage->Size()*nb;
	PBYTE p = pImage->GetBuffer(); 
	for(int i=0; i<size; i++, p++)
	{
		char c1, c2;
		m_ImageStream >> c1 >> c2;
		int n1 = c1 > '9' ? c1-'a'+10 : c1-'0';
		int n2 = c2 > '9' ? c2-'a'+10 : c2-'0';
		*p = BYTE(n1<<4|n2);
	}

	return pImage;
}

bool ZPSImage::WriteFile(const ZImageBase& image, bool intensityscale)
{
	const ZImageBase* pImage = &image;
	ZRGBByteImage cimage;
	ZGrayByteImage gimage;
	if(!is8bit(image) || intensityscale) 
	{
		if(image.isColor())
		{
			image.CopyPixelsTo(cimage, 0, 255);
			pImage = &cimage;
		}
		else
		{
			image.CopyPixelsTo(gimage, 0, 255);
			pImage = &gimage;
		}
	}

	UINT w = image.Width(), h = image.Height();
	float ox=0, oy=0, prw = float(w), prh=float(h);

	PSHead(m_ImageStream, m_filename.c_str(), prw, prh, ox, oy);

	m_ImageStream << "%%Page: 0 0\ngsave\n100 dict begin\n";
	m_ImageStream << prw << ' ' << prh << " scale\n";
	PSpage(*pImage, m_ImageStream);
	m_ImageStream << "end\ngrestore\nshowpage\n%%Trailer\n%%Pages: 0\n%%EOF\n";

	return true;
}
